# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only


import logging

from gi.repository import Gio
from gi.repository import GLib

from gajim.common.util.classes import Singleton

log = logging.getLogger('gajim.c.dbus.file_manager')


class DBusFileManager(metaclass=Singleton):
    '''
    (Partial) implementation of a client for
    https://freedesktop.org/wiki/Specifications/file-manager-interface/
    '''

    def __init__(self) -> None:
        self._proxy: Gio.DBusProxy | None = None
        Gio.DBusProxy.new_for_bus(
            Gio.BusType.SESSION,
            Gio.DBusProxyFlags.DO_NOT_AUTO_START_AT_CONSTRUCTION,
            None,
            'org.freedesktop.FileManager1',
            '/org/freedesktop/FileManager1',
            'org.freedesktop.FileManager1',
            None,
            self._on_proxy_ready,
            )

    def _on_proxy_ready(self, _: None, res: Gio.AsyncResult) -> None:
        try:
            self._proxy = Gio.DBusProxy.new_finish(res)
        except GLib.Error as err:
            log.warning('Couldn’t construct dbus proxy: %s', err.message)

    def is_available(self) -> bool:
        return self._proxy is not None

    def show_items_sync(self,
                        uris: list[str],
                        startup_id: str = '',
                        timeout_ms: int = 500,
                        ) -> bool:
        '''
        Opens parent folder(s), selecting/highlighting the specified
        files/folders.  E.g., for ['file:/home/user/Downloads/xyzzy.foo',
        'file:/home/user/Downloads/Quuxes'], opens ~/Downloads with
        xyzzy.foo and Quuxes items selected.

        Returns True on success.

        Fails if called too early after construction; check `is_available()`.
        '''

        if self._proxy is None:
            log.debug('ShowItems: no dbus proxy')
            return False

        try:
            self._proxy.call_sync(
                'ShowItems',
                GLib.Variant('(ass)', (uris, startup_id)),
                Gio.DBusCallFlags.NONE,
                timeout_ms,
                None,
                )
            return True
        except GLib.Error as err:
            log.warning('ShowItems failed: %s', err.message)
            return False


def init() -> None:
    DBusFileManager()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast

import logging
import sys

from gi.repository import Gtk

from gajim.common import app
from gajim.common import configpaths
from gajim.common import ged
from gajim.common.const import THRESHOLD_OPTIONS
from gajim.common.events import StyleChanged
from gajim.common.events import ThemeUpdate
from gajim.common.i18n import _
from gajim.common.setting_values import BoolSettings

# from gajim.common.util.av import AudioOutputManager
from gajim.common.util.av import AudioInputManager
from gajim.common.util.av import VideoInputManager
from gajim.common.util.uri import open_directory
from gajim.common.util.version import package_version

from gajim.gtk.builder import get_builder
from gajim.gtk.const import Setting
from gajim.gtk.const import SettingKind
from gajim.gtk.const import SettingType
from gajim.gtk.dialogs import ConfirmationDialog
from gajim.gtk.dialogs import DialogButton
from gajim.gtk.preview import PREVIEW_ACTIONS
from gajim.gtk.settings import DropDownSetting
from gajim.gtk.settings import SettingsBox
from gajim.gtk.settings import SettingsDialog
from gajim.gtk.sidebar_switcher import SideBarSwitcher
from gajim.gtk.util.window import get_app_window
from gajim.gtk.util.window import open_window
from gajim.gtk.video_preview import VideoPreview
from gajim.gtk.widgets import GajimAppWindow

log = logging.getLogger("gajim.gtk.preferences")


class Preferences(GajimAppWindow):
    def __init__(self) -> None:
        GajimAppWindow.__init__(
            self,
            name="Preferences",
            title=_("Preferences"),
            default_width=900,
            default_height=650,
            add_window_padding=False,
        )

        self._ui = get_builder("preferences.ui")

        self._video_preview: VideoPreview | None = None
        self._prefs: dict[str, PreferenceBox] = {}

        side_bar_switcher = SideBarSwitcher()
        side_bar_switcher.set_stack(self._ui.stack)
        self._ui.grid.attach(side_bar_switcher, 0, 0, 1, 1)

        self.set_child(self._ui.grid)

        prefs: list[tuple[str, type[PreferenceBox]]] = [
            ("window_behaviour", WindowBehaviour),
            ("plugins", Plugins),
            ("general", General),
            ("chats", Chats),
            ("group_chats", GroupChats),
            ("file_preview", FilePreview),
            ("visual_notifications", VisualNotifications),
            ("sounds", Sounds),
            ("status_message", StatusMessage),
            ("automatic_status", AutomaticStatus),
            ("themes", Themes),
            # ("server", Server),
            ("audio", Audio),
            # ("video", Video),
            ("miscellaneous", Miscellaneous),
            ("advanced", Advanced),
        ]

        self._add_prefs(prefs)
        # self._add_video_preview()

        # self._ui.av_info_bar.set_reveal_child(
        #     not app.is_installed("AV") or sys.platform == "win32"
        # )
        # self._ui.av_info_bar_label.set_text(_("Video calls are not supported"))

    def get_ui(self):
        return self._ui

    def _add_prefs(self, prefs: list[tuple[str, type[PreferenceBox]]]):
        for ui_name, klass in prefs:
            pref_box = cast(Gtk.Grid, getattr(self._ui, ui_name))
            pref = klass(self)  # pyright: ignore
            pref_box.attach(pref, 0, 1, 1, 1)
            self._prefs[ui_name] = pref

    def _add_video_preview(self) -> None:
        self._video_preview = VideoPreview()
        self._ui.video.attach(self._video_preview, 0, 1, 1, 1)

    def get_video_preview(self) -> VideoPreview | None:
        return self._video_preview

    def update_proxy_list(self) -> None:
        miscellaneous = cast(Miscellaneous, self._prefs["miscellaneous"])
        miscellaneous.update_proxy_list()

    def _cleanup(self) -> None:
        del self._video_preview
        self._prefs.clear()


class PreferenceBox(SettingsBox):
    def __init__(self, settings: list[Setting]) -> None:
        SettingsBox.__init__(self, None)
        self.add_css_class("border")
        self.set_selection_mode(Gtk.SelectionMode.NONE)
        self.set_hexpand(True)
        self.set_vexpand(False)
        self.set_valign(Gtk.Align.END)

        for setting in settings:
            self.add_setting(setting)
        self.update_states()


class WindowBehaviour(PreferenceBox):
    def __init__(self, *args: Any) -> None:

        main_window_on_startup_items = {
            "always": _("Always"),
            "never": _("Never"),
            "last_state": _("Restore last state"),
        }

        action_on_close_items = {
            "hide": _("Hide"),
            "minimize": _("Minimize"),
            "quit": _("Quit"),
        }

        settings = [
            Setting(
                SettingKind.DROPDOWN,
                _("Show on Startup"),
                SettingType.CONFIG,
                "show_main_window_on_startup",
                props={"data": main_window_on_startup_items},
                desc=_("Show window when starting Gajim"),
            ),
            Setting(
                SettingKind.DROPDOWN,
                _("Action on Close"),
                SettingType.CONFIG,
                "action_on_close",
                props={"data": action_on_close_items},
                desc=_("Action when closing Gajim’s window"),
            ),
            Setting(
                SettingKind.SWITCH,
                _("Show in Taskbar"),
                SettingType.CONFIG,
                "show_in_taskbar",
                desc=_("Show window in the taskbar"),
                callback=self._on_show_in_taskbar,
            ),
        ]

        PreferenceBox.__init__(self, settings)

    @staticmethod
    def _on_show_in_taskbar(value: bool, *args: Any) -> None:
        app.window.set_skip_taskbar_hint(not value)


class Plugins(PreferenceBox):
    def __init__(self, *args: Any) -> None:

        settings = [
            Setting(
                SettingKind.SWITCH,
                _("Check for updates"),
                SettingType.CONFIG,
                "plugins_update_check",
                desc=_("Check for updates periodically"),
            ),
            Setting(
                SettingKind.SWITCH,
                _("Update automatically"),
                SettingType.CONFIG,
                "plugins_auto_update",
                desc=_("Update plugins automatically"),
                bind="plugins_update_check",
            ),
            Setting(
                SettingKind.SWITCH,
                _("Notify after update"),
                SettingType.CONFIG,
                "plugins_notify_after_update",
                desc=_("Notify me when the automatic update was successful"),
                bind="plugins_auto_update",
            ),
        ]

        PreferenceBox.__init__(self, settings)


class General(PreferenceBox):
    def __init__(self, *args: Any) -> None:

        speller_desc = None
        if not app.is_installed("SPELLING"):
            speller_desc = _("Needs libspelling to be installed")

        settings = [
            Setting(
                SettingKind.SWITCH,
                _("Close with Escape"),
                SettingType.CONFIG,
                "escape_key_closes",
                desc=_("Close a chat by pressing the Escape key"),
            ),
            Setting(
                SettingKind.SWITCH,
                _("Show Send Message Button"),
                SettingType.CONFIG,
                "show_send_message_button",
            ),
            Setting(
                SettingKind.SWITCH,
                _("Show Voice Message Button"),
                SettingType.CONFIG,
                "show_voice_message_button",
            ),
            Setting(
                SettingKind.SWITCH,
                _("Send Messages with Control+Enter"),
                SettingType.CONFIG,
                "send_on_ctrl_enter",
            ),
            Setting(
                SettingKind.SWITCH,
                _("Spell Checking"),
                SettingType.CONFIG,
                "use_speller",
                desc=speller_desc,
                enabled_func=self._speller_available,
            ),
            Setting(
                SettingKind.SWITCH,
                _("Emoji Shortcodes"),
                SettingType.CONFIG,
                "enable_emoji_shortcodes",
                desc=_("Show suggestions for shortcodes, e.g. :+1:"),
            ),
        ]

        PreferenceBox.__init__(self, settings)

    @staticmethod
    def _speller_available() -> bool:
        return app.is_installed("SPELLING")


class Chats(PreferenceBox):
    def __init__(self, *args: Any) -> None:

        settings = [
            Setting(
                SettingKind.SWITCH,
                _("Message Receipts (✔)"),
                SettingType.CONFIG,
                "positive_184_ack",
                desc=_("Add a checkmark to received messages"),
            ),
            Setting(
                SettingKind.SWITCH,
                _("Show Status Changes"),
                SettingType.CONFIG,
                "print_status_in_chats",
                desc=_('For example: "Julia is now online"'),
            ),
        ]

        PreferenceBox.__init__(self, settings)


class GroupChats(PreferenceBox):
    def __init__(self, *args: Any) -> None:

        settings = [
            Setting(
                SettingKind.DROPDOWN,
                _("Default Sync Threshold"),
                SettingType.CONFIG,
                "gc_sync_threshold_public_default",
                desc=_("Default for new public group chats"),
                props={"data": THRESHOLD_OPTIONS},
            ),
            Setting(
                SettingKind.SWITCH,
                _("Direct Messages"),
                SettingType.CONFIG,
                "muc_prefer_direct_msg",
                desc=_("Prefer direct messages in private group chats "),
            ),
            Setting(
                SettingKind.SWITCH,
                _("Sort Participant List by Status"),
                SettingType.CONFIG,
                "sort_by_show_in_muc",
                callback=self._on_sort_by_show_in_muc,
            ),
            Setting(
                SettingKind.SWITCH,
                _("Status Messages in Participants List"),
                SettingType.CONFIG,
                "show_status_msgs_in_roster",
                callback=self._on_show_status_in_roster,
            ),
            Setting(
                SettingKind.SWITCH,
                _("Show Subject"),
                SettingType.CONFIG,
                "show_subject_on_join",
            ),
            Setting(
                SettingKind.SWITCH,
                _("Show Joined / Left"),
                SettingType.CONFIG,
                "gc_print_join_left_default",
                desc=_("Default for new group chats"),
                props={
                    "button-text": _("Reset"),
                    "button-tooltip": _(
                        "Reset all group chats to the current default value"
                    ),
                    "button-style": "destructive-action",
                    "button-callback": self._reset_join_left,
                },
            ),
            Setting(
                SettingKind.SWITCH,
                _("Show Status Changes"),
                SettingType.CONFIG,
                "gc_print_status_default",
                desc=_("Default for new group chats"),
                props={
                    "button-text": _("Reset"),
                    "button-tooltip": _(
                        "Reset all group chats to the current default value"
                    ),
                    "button-style": "destructive-action",
                    "button-callback": self._reset_print_status,
                },
            ),
        ]

        PreferenceBox.__init__(self, settings)

    @staticmethod
    def _on_sort_by_show_in_muc(_value: bool, *args: Any) -> None:
        roster = app.window.get_control().get_group_chat_roster()
        roster.invalidate_sort()

    @staticmethod
    def _on_show_status_in_roster(_value: bool, *args: Any) -> None:
        roster = app.window.get_control().get_group_chat_roster()
        roster.invalidate_sort()

    @staticmethod
    def _reset_join_left(button: Gtk.Button) -> None:
        button.set_sensitive(False)
        app.settings.set_group_chat_settings("print_join_left", None)

    @staticmethod
    def _reset_print_status(button: Gtk.Button) -> None:
        button.set_sensitive(False)
        app.settings.set_group_chat_settings("print_status", None)


class FilePreview(PreferenceBox):
    def __init__(self, *args: Any) -> None:
        sizes = {
            0: _("No automatic preview"),
            262144: "256 KiB",
            524288: "512 KiB",
            1048576: "1 MiB",
            5242880: "5 MiB",
            10485760: "10 MiB",
            26214400: "25 MiB",
        }

        preview_actions = {}
        for action, data in PREVIEW_ACTIONS.items():
            if action == "download":
                continue
            preview_actions[action] = data[0]

        settings = [
            Setting(
                SettingKind.SWITCH,
                _("File Preview"),
                SettingType.CONFIG,
                "enable_file_preview",
                desc=_("Show previews for files"),
            ),
            Setting(
                SettingKind.SPIN,
                _("Preview Size"),
                SettingType.CONFIG,
                "preview_size",
                desc=_("Size of preview images in pixels"),
                bind="enable_file_preview",
                props={"range_": (100, 1000, 1)},
            ),
            Setting(
                SettingKind.DROPDOWN,
                _("File Size Limit"),
                SettingType.CONFIG,
                "preview_max_file_size",
                desc=_("Maximum file size for preview downloads"),
                bind="enable_file_preview",
                props={"data": sizes},
            ),
            Setting(
                SettingKind.SWITCH,
                _("Preview in Public Group Chats"),
                SettingType.CONFIG,
                "preview_anonymous_muc",
                desc=_(
                    "Show previews automatically in public "
                    "group chats (may disclose your data)"
                ),
                bind="enable_file_preview",
            ),
            Setting(
                SettingKind.SWITCH,
                _("Preview all Image URLs"),
                SettingType.CONFIG,
                "preview_allow_all_images",
                desc=_("Show previews for any URLs containing images (may be unsafe)"),
                bind="enable_file_preview",
            ),
            Setting(
                SettingKind.DROPDOWN,
                _("Left Click Action"),
                SettingType.CONFIG,
                "preview_leftclick_action",
                desc=_("Action for left-clicking a preview"),
                bind="enable_file_preview",
                props={"data": preview_actions},
            ),
            Setting(
                SettingKind.SWITCH,
                _("HTTPS Verification"),
                SettingType.CONFIG,
                "preview_verify_https",
                desc=_(
                    "Whether to check for a valid certificate before "
                    "downloading (not safe to disable)"
                ),
                bind="enable_file_preview",
            ),
        ]

        PreferenceBox.__init__(self, settings)


class VisualNotifications(PreferenceBox):
    def __init__(self, *args: Any) -> None:

        settings = [
            Setting(
                SettingKind.SWITCH,
                _("Notification Area Icon"),
                SettingType.CONFIG,
                "show_trayicon",
            ),
            Setting(
                SettingKind.NOTIFICATIONS,
                _("Show Notifications"),
                SettingType.DIALOG,
                props={"dialog": NotificationsDialog},
            ),
        ]

        PreferenceBox.__init__(self, settings)


class NotificationsDialog(SettingsDialog):
    def __init__(self, account: str, parent: Gtk.Window) -> None:

        settings = [
            Setting(
                SettingKind.SWITCH,
                _("Show Notifications"),
                SettingType.CONFIG,
                "show_notifications",
            ),
            Setting(
                SettingKind.SWITCH,
                _("Notifications When Away"),
                SettingType.CONFIG,
                "show_notifications_away",
                desc=_("Show notifications even if you are Away, Busy, etc."),
                bind="show_notifications",
            ),
        ]

        SettingsDialog.__init__(
            self, parent, _("Notifications"), Gtk.DialogFlags.MODAL, settings, account
        )


class Sounds(PreferenceBox):
    def __init__(self, *args: Any) -> None:

        settings = [
            Setting(
                SettingKind.SWITCH,
                _("Play Sounds"),
                SettingType.CONFIG,
                "sounds_on",
                desc=_("Play sounds to notify about events"),
                props={
                    "button-icon-name": "preferences-system-symbolic",
                    "button-callback": self._on_manage_sounds,
                },
            ),
            Setting(
                SettingKind.SWITCH,
                _("Sounds When Away"),
                SettingType.CONFIG,
                "sounddnd",
                desc=_("Play sounds even when you are Away, Busy, etc."),
                bind="sounds_on",
            ),
        ]

        PreferenceBox.__init__(self, settings)

    @staticmethod
    def _on_manage_sounds(_button: Gtk.Button) -> None:
        window = get_app_window("Preferences")
        assert window is not None
        open_window("ManageSounds", transient_for=window.window)


class StatusMessage(PreferenceBox):
    def __init__(self, *args: Any) -> None:

        settings = [
            Setting(
                SettingKind.SWITCH,
                _("Sign In"),
                SettingType.CONFIG,
                "ask_online_status",
            ),
        ]

        PreferenceBox.__init__(self, settings)


class AutomaticStatus(PreferenceBox):
    def __init__(self, *args: Any) -> None:

        settings = [
            Setting(
                SettingKind.AUTO_AWAY,
                _("Auto Away"),
                SettingType.DIALOG,
                desc=_(
                    'Change your status to "Away" after a certain ' "amount of time"
                ),
                props={"dialog": AutoAwayDialog},
            ),
            Setting(
                SettingKind.AUTO_EXTENDED_AWAY,
                _("Auto Not Available"),
                SettingType.DIALOG,
                desc=_(
                    'Change your status to "Not Available" after a '
                    "certain amount of time"
                ),
                props={"dialog": AutoExtendedAwayDialog},
            ),
        ]

        PreferenceBox.__init__(self, settings)

    @staticmethod
    def _get_auto_away() -> bool:
        return app.settings.get("autoaway")

    @staticmethod
    def _get_auto_xa() -> bool:
        return app.settings.get("autoxa")


class AutoAwayDialog(SettingsDialog):
    def __init__(self, account: str, parent: Gtk.Window) -> None:

        settings = [
            Setting(SettingKind.SWITCH, _("Auto Away"), SettingType.CONFIG, "autoaway"),
            Setting(
                SettingKind.SPIN,
                _("Time Until Away"),
                SettingType.CONFIG,
                "autoawaytime",
                desc=_("Minutes until your status gets changed"),
                props={"range_": (1, 720, 1)},
                bind="autoaway",
            ),
            Setting(
                SettingKind.ENTRY,
                _("Status Message"),
                SettingType.CONFIG,
                "autoaway_message",
                bind="autoaway",
            ),
        ]

        SettingsDialog.__init__(
            self,
            parent,
            _("Auto Away Settings"),
            Gtk.DialogFlags.MODAL,
            settings,
            account,
        )


class AutoExtendedAwayDialog(SettingsDialog):
    def __init__(self, account: str, parent: Gtk.Window) -> None:

        settings = [
            Setting(
                SettingKind.SWITCH,
                _("Auto Not Available"),
                SettingType.CONFIG,
                "autoxa",
            ),
            Setting(
                SettingKind.SPIN,
                _("Time Until Not Available"),
                SettingType.CONFIG,
                "autoxatime",
                desc=_("Minutes until your status gets changed"),
                props={"range_": (1, 720, 1)},
                bind="autoxa",
            ),
            Setting(
                SettingKind.ENTRY,
                _("Status Message"),
                SettingType.CONFIG,
                "autoxa_message",
                bind="autoxa",
            ),
        ]

        SettingsDialog.__init__(
            self,
            parent,
            _("Auto Extended Away Settings"),
            Gtk.DialogFlags.MODAL,
            settings,
            account,
        )


class Themes(PreferenceBox):
    def __init__(self, *args: Any) -> None:

        theme_items = self._get_theme_items()

        dark_theme_items = {
            0: _("Disabled"),
            1: _("Enabled"),
            2: _("System"),
        }

        settings = [
            Setting(
                SettingKind.SPIN,
                _("User Interface Font Size"),
                SettingType.CONFIG,
                "app_font_size",
                props={"range_": (1.0, 1.5, 0.125)},
                callback=self._on_app_font_size_changed,
            ),
            Setting(
                SettingKind.DROPDOWN,
                _("Dark Theme"),
                SettingType.CONFIG,
                "dark_theme",
                props={"data": dark_theme_items},
                callback=self._on_dark_theme,
            ),
            Setting(
                SettingKind.DROPDOWN,
                _("Theme"),
                SettingType.CONFIG,
                "roster_theme",
                name="roster_theme",
                props={
                    "data": theme_items,
                    "button-icon-name": "preferences-system-symbolic",
                    "button-callback": self._on_edit_themes,
                },
                callback=self._on_theme_changed,
            ),
        ]

        PreferenceBox.__init__(self, settings)

        app.ged.register_event_handler("theme-update", ged.GUI1, self._on_theme_update)

    def do_unroot(self) -> None:
        app.ged.remove_event_handler("theme-update", ged.GUI1, self._on_theme_update)
        PreferenceBox.do_unroot(self)

    @staticmethod
    def _on_app_font_size_changed(_value: float, *args: Any) -> None:
        app.css_config.apply_app_font_size()

    @staticmethod
    def _get_theme_items() -> list[str]:
        return ["default", *app.css_config.themes]

    def _on_theme_update(self, *args: Any) -> None:
        dropdown_row = cast(DropDownSetting, self.get_setting("roster_theme"))
        dropdown_row.update_entries(self._get_theme_items())
        dropdown_row.select_key(app.settings.get("roster_theme"))

    def _on_edit_themes(self, _button: Gtk.Button) -> None:
        window = get_app_window("Preferences")
        assert window is not None
        open_window("Themes", transient=window.window)

    def _on_theme_changed(self, value: str, *args: Any) -> None:
        app.ged.remove_event_handler("theme-update", ged.GUI1, self._on_theme_update)

        app.css_config.change_theme(value)
        app.ged.raise_event(ThemeUpdate())

        app.ged.register_event_handler("theme-update", ged.GUI1, self._on_theme_update)

    @staticmethod
    def _on_dark_theme(value: str, *args: Any) -> None:
        app.css_config.set_dark_theme(int(value))
        app.css_config.reload_css()
        app.ged.raise_event(StyleChanged())


class Server(PreferenceBox):
    def __init__(self, *args: Any) -> None:

        settings = [
            Setting(
                SettingKind.USE_STUN_SERVER,
                _("Use STUN Server"),
                SettingType.DIALOG,
                desc=_("Helps to establish calls through firewalls"),
                props={"dialog": StunServerDialog},
            ),
        ]

        PreferenceBox.__init__(self, settings)

        self.set_sensitive(app.is_installed("AV"))


class StunServerDialog(SettingsDialog):
    def __init__(self, account: str, parent: Gtk.Window) -> None:

        settings = [
            Setting(
                SettingKind.SWITCH,
                _("Use STUN Server"),
                SettingType.CONFIG,
                "use_stun_server",
            ),
            Setting(
                SettingKind.ENTRY,
                _("STUN Server"),
                SettingType.CONFIG,
                "stun_server",
                bind="use_stun_server",
            ),
        ]

        SettingsDialog.__init__(
            self,
            parent,
            _("STUN Server Settings"),
            Gtk.DialogFlags.MODAL,
            settings,
            account,
        )


class Audio(PreferenceBox):
    def __init__(self, *args: Any) -> None:

        deps_installed = app.is_installed("GST")

        audio_input_devices = {}
        # audio_output_devices = {}
        if deps_installed:
            audio_input_devices = AudioInputManager().get_devices()
            # audio_output_devices = AudioOutputManager().get_devices()

        audio_input_items = self._create_av_combo_items(audio_input_devices)
        # audio_output_items = self._create_av_combo_items(audio_output_devices)

        settings = [
            Setting(
                SettingKind.DROPDOWN,
                _("Audio Input Device"),
                SettingType.CONFIG,
                "audio_input_device",
                desc=_("Select your audio input (e.g. microphone)"),
                props={"data": audio_input_items},
            ),
            # Setting(
            #     SettingKind.DROPDOWN,
            #     _("Audio Output Device"),
            #     SettingType.CONFIG,
            #     "audio_output_device",
            #     desc=_("Select an audio output (e.g. speakers, headphones)"),
            #     props={"data": audio_output_items},
            # ),
        ]

        PreferenceBox.__init__(self, settings)

        self.set_sensitive(deps_installed)

    @staticmethod
    def _create_av_combo_items(items_dict: dict[str, str]) -> dict[str, str]:
        items = enumerate(
            sorted(
                items_dict.items(),
                key=lambda x: "" if x[1].startswith("auto") else x[0].lower(),
            )
        )
        combo_items: dict[str, str] = {}
        for _index, (name, value) in items:
            combo_items[value] = name
        return combo_items


class Video(PreferenceBox):
    def __init__(self, *args: Any) -> None:

        deps_installed = app.is_installed("AV")

        video_input_devices = {}
        if deps_installed:
            video_input_devices = VideoInputManager().get_devices()

        video_input_items = self._create_av_combo_items(video_input_devices)

        video_framerates = {
            "": _("Default"),
            "15/1": "15 fps",
            "10/1": "10 fps",
            "5/1": "5 fps",
            "5/2": "2.5 fps",
        }

        video_sizes = {
            "": _("Default"),
            "800x600": "800x600",
            "640x480": "640x480",
            "320x240": "320x240",
        }

        settings = [
            Setting(
                SettingKind.DROPDOWN,
                _("Video Input Device"),
                SettingType.CONFIG,
                "video_input_device",
                props={"data": video_input_items},
                desc=_("Select your video input device (e.g. webcam, screen capture)"),
                callback=self._on_video_input_changed,
            ),
            Setting(
                SettingKind.DROPDOWN,
                _("Video Framerate"),
                SettingType.CONFIG,
                "video_framerate",
                props={"data": video_framerates},
            ),
            Setting(
                SettingKind.DROPDOWN,
                _("Video Resolution"),
                SettingType.CONFIG,
                "video_size",
                props={"data": video_sizes},
            ),
            Setting(
                SettingKind.SWITCH,
                _("Show My Video Stream"),
                SettingType.CONFIG,
                "video_see_self",
                desc=_("Show your own video stream in calls"),
            ),
            Setting(
                SettingKind.SWITCH,
                _("Live Preview"),
                SettingType.VALUE,
                desc=_("Show a live preview to test your video source"),
                callback=self._toggle_live_preview,
            ),
        ]

        PreferenceBox.__init__(self, settings)

        self.set_sensitive(deps_installed)

    @staticmethod
    def _on_video_input_changed(_value: str, *args: Any) -> None:
        window = get_app_window("Preferences")
        assert window is not None
        preview = window.get_video_preview()
        if preview is None or not preview.is_active:
            # changed signal gets triggered when we fill the combobox
            return
        preview.refresh()

    @staticmethod
    def _toggle_live_preview(value: bool, *args: Any) -> None:
        window = get_app_window("Preferences")
        assert window is not None
        preview = window.get_video_preview()
        if preview is not None:
            preview.toggle_preview(value)

    @staticmethod
    def _create_av_combo_items(items_dict: dict[str, str]) -> dict[str, str]:
        items = enumerate(
            sorted(
                items_dict.items(),
                key=lambda x: "" if x[1].startswith("auto") else x[0].lower(),
            )
        )
        combo_items: dict[str, str] = {}
        for _index, (name, value) in items:
            combo_items[value] = name
        return combo_items


class Miscellaneous(PreferenceBox):
    def __init__(self, pref_window: Preferences) -> None:
        self._hints_list: list[BoolSettings] = [
            "show_help_start_chat",
        ]

        settings = [
            Setting(
                SettingKind.DROPDOWN,
                _("Global Proxy"),
                SettingType.CONFIG,
                "global_proxy",
                name="global_proxy",
                props={
                    "data": self._get_proxies(),
                    "button-icon-name": "preferences-system-symbolic",
                    "button-callback": self._on_proxy_edit,
                },
            ),
            Setting(
                SettingKind.SWITCH,
                _("Use System Keyring"),
                SettingType.CONFIG,
                "use_keyring",
                desc=_("Use your system’s keyring to store passwords"),
            ),
        ]

        if sys.platform not in ("win32", "darwin") and package_version(
            "keyring>=23.8.1"
        ):
            settings.append(
                Setting(
                    SettingKind.SWITCH,
                    _("Enable KeepassXC Integration"),
                    SettingType.CONFIG,
                    "enable_keepassxc_integration",
                )
            )

        if sys.platform in ("win32", "darwin"):
            settings.append(
                Setting(
                    SettingKind.SWITCH,
                    _("Check For Updates"),
                    SettingType.CONFIG,
                    "check_for_update",
                    desc=_("Check for Gajim updates periodically"),
                )
            )

        PreferenceBox.__init__(self, settings)

        reset_button = pref_window.get_ui().reset_button
        self._connect(reset_button, "clicked", self._on_reset_hints)
        reset_button.set_sensitive(self._check_hints_reset())

        purge_history_button = pref_window.get_ui().purge_history_button
        self._connect(purge_history_button, "clicked", self._on_purge_history_clicked)

    @staticmethod
    def _get_proxies() -> dict[str, str]:
        proxies = {"": _("System")}
        proxies.update({proxy: proxy for proxy in app.settings.get_proxies()})
        proxies["no-proxy"] = _("No Proxy")
        return proxies

    @staticmethod
    def _on_proxy_edit(*args: Any) -> None:
        window = get_app_window("Preferences")
        assert window is not None
        open_window("ManageProxies", transient_for=window.window)

    def update_proxy_list(self) -> None:
        dropdown_row = cast(DropDownSetting, self.get_setting("global_proxy"))
        dropdown_row.update_entries(self._get_proxies())

    def _check_hints_reset(self) -> bool:
        return any(app.settings.get(hint) is False for hint in self._hints_list)

    def _on_reset_hints(self, button: Gtk.Button) -> None:
        for hint in self._hints_list:
            app.settings.set(hint, True)
        button.set_sensitive(False)

    @staticmethod
    def _on_purge_history_clicked(button: Gtk.Button) -> None:
        def _purge() -> None:
            button.set_sensitive(False)
            app.storage.archive.remove_all_history()
            app.window.quit()

        ConfirmationDialog(
            _("Purge all Chat History?"),
            _(
                "Do you really want to remove all chat messages from Gajim?\n"
                "Warning: This can’t be undone!\n"
                "Gajim will quit afterwards."
            ),
            [
                DialogButton.make("Cancel"),
                DialogButton.make("Remove", text=_("_Purge"), callback=_purge),
            ],
        )


class Advanced(PreferenceBox):
    def __init__(self, pref_window: Preferences) -> None:

        settings = [
            Setting(
                SettingKind.SWITCH,
                _("Debug Logging"),
                SettingType.VALUE,
                app.get_debug_mode(),
                props={
                    "button-icon-name": "folder-symbolic",
                    "button-callback": self._on_open_debug_logs,
                },
                callback=self._on_debug_logging,
            ),
            Setting(
                SettingKind.SWITCH,
                _("D-Bus Interface"),
                SettingType.CONFIG,
                "remote_control",
                desc=_(
                    "Allow Gajim to broadcast useful information via "
                    "D-Bus. It also allows other applications to "
                    "control Gajim remotely."
                ),
            ),
        ]

        PreferenceBox.__init__(self, settings)

        pref_window.get_ui().ace_button.connect(
            "clicked", self._on_advanced_config_editor
        )

    @staticmethod
    def _on_debug_logging(value: bool, *args: Any) -> None:
        app.set_debug_mode(value)

    @staticmethod
    def _on_open_debug_logs(*args: Any) -> None:
        open_directory(configpaths.get("DEBUG"))

    @staticmethod
    def _on_advanced_config_editor(*args: Any) -> None:
        window = get_app_window("Preferences")
        assert window is not None
        open_window("AdvancedConfig", transient_for=window.window)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from gi.repository import Gtk

from gajim.common import app
from gajim.common.i18n import _
from gajim.common.styling import ParsingResult
from gajim.common.styling import PlainBlock
from gajim.common.styling import PreBlock
from gajim.common.styling import process
from gajim.common.styling import QuoteBlock

from gajim.gtk.const import MAX_MESSAGE_LENGTH
from gajim.gtk.conversation.code_widget import CodeWidget
from gajim.gtk.conversation.plain_widget import PlainWidget
from gajim.gtk.conversation.quote_widget import QuoteWidget
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import container_remove_all
from gajim.gtk.util.misc import iterate_children
from gajim.gtk.widgets import GajimAppWindow

ContentT = ParsingResult | QuoteBlock


class MessageWidget(Gtk.Box, SignalManager):
    def __init__(self, account: str, selectable: bool = True) -> None:
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL)
        SignalManager.__init__(self)

        self._account = account
        self._selectable = selectable

        self._content: ContentT | None = None
        self._original_text = ""
        self._action_phrase_text = ""

    def do_unroot(self) -> None:
        self._disconnect_all()
        Gtk.Box.do_unroot(self)
        app.check_finalize(self)

    def get_content(self) -> ContentT | None:
        return self._content

    def get_text(self) -> str:
        if self._content is not None:
            return self._original_text
        return self._action_phrase_text

    def set_selectable(self, selectable: bool) -> None:
        for widget in iterate_children(self):
            if isinstance(widget, PlainWidget):
                widget.set_selectable(selectable)

    def add_with_styling(self, text: str, nickname: str | None = None) -> None:

        self._original_text = text
        if len(text) > MAX_MESSAGE_LENGTH:
            text = text[:MAX_MESSAGE_LENGTH]

        if text.startswith("/me ") and nickname is not None:
            self._add_action_phrase(text, nickname)
            if len(self._original_text) > MAX_MESSAGE_LENGTH:
                self._add_read_more_button(self._original_text)
            return

        result = process(text)
        self.add_content(result)
        if len(self._original_text) > MAX_MESSAGE_LENGTH:
            self._add_read_more_button(self._original_text)

    def _add_action_phrase(self, text: str, nickname: str):
        self.clear()
        widget = PlainWidget(self._account, self._selectable)
        widget.add_action_phrase(text, nickname)
        self.append(widget)

        self._action_phrase_text = text.replace("/me", f"* {nickname}", 1)

    def add_content(self, content: ContentT) -> None:
        self.clear()
        self._content = content
        for block in content.blocks:
            if isinstance(block, PlainBlock):
                widget = PlainWidget(self._account, self._selectable)
                widget.add_content(block)
                self.append(widget)
                continue

            if isinstance(block, PreBlock):
                widget = CodeWidget(self._account)
                widget.add_content(block)
                self.append(widget)
                continue

            if isinstance(block, QuoteBlock):
                message_widget = MessageWidget(self._account, self._selectable)
                message_widget.add_content(block)
                widget = QuoteWidget(self._account)
                widget.append(message_widget)
                self.append(widget)
                continue

    def _add_read_more_button(self, text: str) -> None:
        link_button = Gtk.LinkButton(label=_("[read more]"))
        link_button.set_halign(Gtk.Align.START)
        self._connect(link_button, "activate-link", self._on_read_more, text)
        self.append(link_button)

    def _on_read_more(self, _button: Gtk.LinkButton, text: str) -> bool:
        FullMessageWindow(text)
        return True

    def clear(self) -> None:
        container_remove_all(self)


class FullMessageWindow(GajimAppWindow):
    def __init__(self, text: str) -> None:
        GajimAppWindow.__init__(
            self,
            name="FullMessageWindow",
            title=_("Full Message View"),
            default_width=800,
            default_height=800,
        )

        textview = Gtk.TextView(
            editable=False,
            wrap_mode=Gtk.WrapMode.WORD_CHAR,
        )
        textview.get_buffer().set_text(text)

        scrolled = Gtk.ScrolledWindow(child=textview, hexpand=True)

        box = Gtk.Box()
        box.append(scrolled)

        self.set_child(box)
        self.show()

    def _cleanup(self) -> None:
        pass



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from gi.repository import Gio
from gi.repository import GLib
from gi.repository import Gtk

from gajim.common import app


class MainMenuButton(Gtk.MenuButton):
    def __init__(self) -> None:
        Gtk.MenuButton.__init__(
            self,
            halign=Gtk.Align.CENTER,
            margin_top=12,
            icon_name="open-menu-symbolic",
        )

        menu_model = app.app.get_menubar()
        assert menu_model is not None

        menu = Gtk.PopoverMenu.new_from_model_full(
            menu_model, Gtk.PopoverMenuFlags.NESTED
        )

        self.set_popover(menu)

        menu_toggle_action = app.window.lookup_action("toggle-menu-bar")
        assert menu_toggle_action is not None
        menu_toggle_action.connect("activate", self._on_menu_toggle_action)

        self.set_visible(not app.settings.get_app_setting("show_main_menu"))

    def _on_menu_toggle_action(
        self, _action: Gio.SimpleAction, _param: GLib.Variant | None
    ) -> None:
        self.set_visible(not app.window.get_show_menubar())



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import logging

from gi.repository import Gtk
from nbxmpp.errors import MalformedStanzaError
from nbxmpp.errors import StanzaError
from nbxmpp.task import Task

from gajim.common import app
from gajim.common.i18n import _

from gajim.gtk.builder import get_builder
from gajim.gtk.dialogs import ConfirmationDialog
from gajim.gtk.dialogs import DialogButton
from gajim.gtk.dialogs import SimpleDialog
from gajim.gtk.util.misc import ensure_not_destroyed
from gajim.gtk.widgets import GajimAppWindow

log = logging.getLogger("gajim.gtk.mam_preferences")


class MamPreferences(GajimAppWindow):
    def __init__(self, account: str) -> None:
        GajimAppWindow.__init__(
            self,
            name="MamPreferences",
            title=_("Archiving Preferences for %s") % account,
        )

        self.account = account
        self._client = app.get_client(account)
        self._destroyed = False

        self._ui = get_builder("mam_preferences.ui")
        self.set_child(self._ui.mam_box)

        default_store = Gtk.ListStore(str, str)
        default_store.append([_("Always"), "always"])
        default_store.append([_("Contact List"), "roster"])
        default_store.append([_("Never"), "never"])
        self._ui.default_combo.set_model(default_store)

        self._preferences_store = Gtk.ListStore(str, bool)
        self._ui.pref_view.set_model(self._preferences_store)

        self._connect(self._ui.jid_cell_renderer, "edited", self._jid_edited)
        self._connect(self._ui.pref_toggle_cell_renderer, "toggled", self._pref_toggled)
        self._connect(self._ui.add, "clicked", self._on_add)
        self._connect(self._ui.remove, "clicked", self._on_remove)
        self._connect(self._ui.save_button, "clicked", self._on_save)

        self._spinner = Gtk.Spinner()
        self._ui.overlay.add_overlay(self._spinner)

        self._set_mam_box_state(False)
        self._activate_spinner()

        self._client.get_module("MAM").request_preferences(
            callback=self._mam_prefs_received
        )

    def _cleanup(self) -> None:
        self._destroyed = True

    @ensure_not_destroyed
    def _mam_prefs_received(self, task: Task) -> None:
        try:
            result = task.finish()
        except (StanzaError, MalformedStanzaError) as error:
            self._on_error(error.get_text())
            return

        self._disable_spinner()
        self._set_mam_box_state(True)

        self._ui.default_combo.set_active_id(result.default)

        self._preferences_store.clear()
        for jid in result.always:
            self._preferences_store.append([str(jid), True])

        for jid in result.never:
            self._preferences_store.append([str(jid), False])

    @ensure_not_destroyed
    def _mam_prefs_saved(self, task: Task) -> None:
        try:
            task.finish()
        except StanzaError as error:
            self._on_error(error.get_text())
            return

        self._disable_spinner()

        def _on_ok():
            self.close()

        ConfirmationDialog(
            _("Archiving Preferences Saved"),
            _("Your archiving preferences have successfully been saved."),
            [DialogButton.make("OK", callback=_on_ok)],
        )

    def _on_error(self, error: str) -> None:
        self._disable_spinner()

        SimpleDialog(
            _("Archiving Preferences Error"), _("Error received: {}").format(error)
        )

        self._set_mam_box_state(True)

    def _set_mam_box_state(self, state: bool) -> None:
        self._ui.mam_box.set_sensitive(state)

    def _jid_edited(
        self, _renderer: Gtk.CellRendererText, path: str, new_text: str
    ) -> None:

        iter_ = self._preferences_store.get_iter(path)
        self._preferences_store.set_value(iter_, 0, new_text)

    def _pref_toggled(self, _renderer: Gtk.CellRendererToggle, path: str):

        iter_ = self._preferences_store.get_iter(path)
        current_value = self._preferences_store[iter_][1]
        self._preferences_store.set_value(iter_, 1, not current_value)

    def _on_add(self, _button: Gtk.Button) -> None:
        self._preferences_store.append(["", False])

    def _on_remove(self, _button: Gtk.Button) -> None:
        rows = self._ui.pref_view.get_selection().get_selected_rows()
        assert rows is not None
        mod, paths = rows
        for path in paths:
            iter_ = mod.get_iter(path)
            self._preferences_store.remove(iter_)

    def _on_save(self, _button: Gtk.Button) -> None:
        self._activate_spinner()
        self._set_mam_box_state(False)
        always: list[str] = []
        never: list[str] = []
        default = self._ui.default_combo.get_active_id()
        for item in self._preferences_store:
            jid, archive = item
            if archive:
                always.append(jid)
            else:
                never.append(jid)
        self._client.get_module("MAM").set_preferences(
            default, always, never, callback=self._mam_prefs_saved
        )

    def _activate_spinner(self) -> None:
        self._spinner.set_visible(True)
        self._spinner.start()

    def _disable_spinner(self) -> None:
        self._spinner.set_visible(False)
        self._spinner.stop()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from typing import NamedTuple

import binascii
import hashlib
import logging
import mimetypes
from pathlib import Path
from urllib.parse import ParseResult
from urllib.parse import unquote
from urllib.parse import urlparse

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import algorithms
from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers.modes import GCM
from gi.repository import GdkPixbuf
from gi.repository import Gio
from gi.repository import GLib
from PIL import Image
from PIL import ImageFile

from gajim.common.helpers import sanitize_filename
from gajim.common.i18n import _
from gajim.common.i18n import p_

log = logging.getLogger('gajim.c.util.preview')

ImageFile.LOAD_TRUNCATED_IMAGES = True


class Coords(NamedTuple):
    location: str
    lat: str
    lon: str


def parse_fragment(fragment_string: str) -> tuple[bytes, bytes]:
    if not fragment_string:
        raise ValueError('Invalid fragment')

    fragment = binascii.unhexlify(fragment_string)
    size = len(fragment)
    # Clients started out with using a 16 byte IV but long term
    # want to switch to the more performant 12 byte IV
    # We have to support both
    if size == 48:
        key = fragment[16:]
        iv = fragment[:16]
    elif size == 44:
        key = fragment[12:]
        iv = fragment[:12]
    else:
        raise ValueError('Invalid fragment size: %s' % size)

    return key, iv


def get_image_paths(uri: str,
                    urlparts: ParseResult,
                    size: int,
                    orig_dir: Path,
                    thumb_dir: Path) -> tuple[Path, Path]:

    path = Path(unquote(urlparts.path))
    web_stem = path.stem
    extension = path.suffix

    web_stem = sanitize_filename(web_stem)

    name_hash = hashlib.sha1(str(uri).encode()).hexdigest()

    orig_filename = f'{web_stem}_{name_hash}{extension}'

    thumb_filename = f'{web_stem}_{name_hash}_thumb_{size}.png'

    orig_path = orig_dir / orig_filename
    thumb_path = thumb_dir / thumb_filename
    return orig_path, thumb_path


def split_geo_uri(uri: str) -> Coords:
    # Example:
    # geo:37.786971,-122.399677,122.3;CRS=epsg:32718;U=20;mapcolors=abc
    # Assumption is all coordinates are CRS=WGS-84

    # Remove "geo:"
    coords = uri[4:]

    # Remove arguments
    if ';' in coords:
        coords, _ = coords.split(';', maxsplit=1)

    # Split coords
    coords = coords.split(',')
    if len(coords) not in (2, 3):
        raise ValueError('Invalid geo uri: invalid coord count')

    # Remoove coord-c (altitude)
    if len(coords) == 3:
        coords.pop(2)

    lat, lon = coords
    if float(lat) < -90 or float(lat) > 90:
        raise ValueError('Invalid geo_uri: invalid latitude %s' % lat)

    if float(lon) < -180 or float(lon) > 180:
        raise ValueError('Invalid geo_uri: invalid longitude %s' % lon)

    location = ','.join(coords)
    return Coords(location=location, lat=lat, lon=lon)


def format_geo_coords(lat: float, lon: float) -> str:
    def fmt(f: float) -> str:
        i = int(round(f * 3600.0))
        seconds = i % 60
        i //= 60
        minutes = i % 60
        i //= 60
        degrees = i
        return '%d°%02d′%02d′′' % (degrees, minutes, seconds)
    if lat >= 0:
        slat = p_('positive latitude', '%sN') % fmt(lat)
    else:
        slat = p_('negative latitude', '%sS') % fmt(-lat)
    if lon >= 0:
        slon = p_('positive longitude', '%sE') % fmt(lon)
    else:
        slon = p_('negative longitude', '%sW') % fmt(-lon)
    return f'{slat} {slon}'


def filename_from_uri(uri: str) -> str:
    urlparts = urlparse(unquote(uri))
    path = Path(urlparts.path)
    return path.name


def aes_decrypt(key: bytes, iv: bytes, payload: bytes) -> bytes:
    # Use AES256 GCM with the given key and iv to decrypt the payload
    data = payload[:-16]
    tag = payload[-16:]
    decryptor = Cipher(algorithms.AES(key),
                       GCM(iv, tag=tag),
                       backend=default_backend()).decryptor()
    return decryptor.update(data) + decryptor.finalize()


def contains_audio_streams(file_path: Path) -> bool:
    # Check if it is really an audio file

    from gi.repository import GstPbutils

    has_audio = False
    discoverer = GstPbutils.Discoverer()
    try:
        info = discoverer.discover_uri(file_path.as_uri())
        has_audio = bool(info.get_audio_streams())
    except GLib.Error as err:
        log.error('Error while reading %s: %s', str(file_path), err)
        return False
    if not has_audio:
        log.warning('File does not contain audio stream: %s', str(file_path))
    return has_audio


def get_previewable_mime_types() -> set[str]:
    previewable_mime_types: set[str] = set()
    for fmt in GdkPixbuf.Pixbuf.get_formats():
        mime_types = fmt.get_mime_types()
        if mime_types is None:
            continue
        for mime_type in mime_types:
            previewable_mime_types.add(mime_type.lower())

    Image.init()
    for mime_type in Image.MIME.values():
        previewable_mime_types.add(mime_type.lower())

    return set(filter(
        lambda mime_type: mime_type.startswith('image'),
        previewable_mime_types
    ))


def guess_mime_type(file_path: Path | str,
                    data: bytes | None = None
                    ) -> str:
    file_path = str(file_path)

    # The mimetypes module maps extensions to mime types
    # it does no guessing based on file content
    mime_type, _ = mimetypes.guess_type(file_path)
    if mime_type is None:
        # Gio does also guess based on file content
        extension, _ = Gio.content_type_guess(file_path, data)
        mime_type = Gio.content_type_get_mime_type(extension)

    log.debug('Guessed MIME type: %s', mime_type)
    return mime_type or ''


def guess_simple_file_type(file_path: str,
                           data: bytes | None = None
                           ) -> tuple[Gio.Icon, str]:
    mime_type = guess_mime_type(file_path, data)
    if mime_type == 'application/octet-stream':
        mime_type = ''
    icon = get_icon_for_mime_type(mime_type)
    if mime_type.startswith('audio/'):
        return icon, _('Audio File')
    if mime_type.startswith('image/'):
        return icon, _('Image')
    if mime_type.startswith('video/'):
        return icon, _('Video')
    if mime_type.startswith('text/'):
        return icon, _('Text File')
    return icon, _('File')


def get_icon_for_mime_type(mime_type: str) -> Gio.Icon:
    if mime_type == '':
        return Gio.Icon.new_for_string('mail-attachment')
    return Gio.content_type_get_icon(mime_type)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import Literal
from typing import TYPE_CHECKING

import logging

from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk
from nbxmpp import JID

from gajim.common import app
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.modules.contacts import GroupchatParticipant

from gajim.gtk.builder import get_builder
from gajim.gtk.chat_filter import ChatFilter
from gajim.gtk.chat_list import ChatList
from gajim.gtk.chat_list_stack import ChatListStack
from gajim.gtk.chat_stack import ChatStack
from gajim.gtk.menus import get_start_chat_button_menu
from gajim.gtk.search_view import SearchView
from gajim.gtk.util.classes import SignalManager

if TYPE_CHECKING:
    from gajim.gtk.control import ChatControl

log = logging.getLogger("gajim.gtk.chat_page")


class ChatPage(Gtk.Box, SignalManager):

    __gsignals__ = {
        "chat-selected": (GObject.SignalFlags.RUN_LAST, None, (str, str, str)),
    }

    def __init__(self):
        Gtk.Box.__init__(self)
        SignalManager.__init__(self)

        self._ui = get_builder("chat_paned.ui")
        self.append(self._ui.paned)

        self._chat_stack = ChatStack()
        self._ui.right_grid.attach(self._chat_stack, 0, 0, 1, 1)

        self._chat_control = self._chat_stack.get_chat_control()

        self._search_view = SearchView()
        self._search_view.connect("hide-search", self._on_search_hide)

        self._search_revealer = Gtk.Revealer()
        self._search_revealer.set_reveal_child(False)
        self._search_revealer.set_hexpand(False)
        self._search_revealer.set_hexpand_set(True)
        self._search_revealer.set_transition_type(Gtk.RevealerTransitionType.SLIDE_LEFT)
        self._search_revealer.set_child(self._search_view)
        self._ui.right_grid.attach(self._search_revealer, 1, 0, 1, 1)

        self._restore_occupants_list = False

        section_hover_controller = Gtk.EventControllerMotion()
        self._connect(section_hover_controller, "enter", self._on_section_label_enter)
        self._connect(section_hover_controller, "leave", self._on_section_label_leave)
        self._ui.section_label_eventbox.add_controller(section_hover_controller)

        self._chat_filter = ChatFilter()
        self._chat_filter.insert_after(self._ui.controls_box, self._ui.search_entry)

        self._chat_list_stack = ChatListStack(self._chat_filter, self._ui.search_entry)
        self._connect(self._chat_list_stack, "chat-selected", self._on_chat_selected)
        self._connect(
            self._chat_list_stack, "chat-unselected", self._on_chat_unselected
        )
        self._connect(self._chat_list_stack, "chat-removed", self._on_chat_removed)
        self._connect(
            self._chat_list_stack,
            "notify::visible-child-name",
            self._on_chat_list_changed,
        )
        self._ui.chat_list_scrolled.set_child(self._chat_list_stack)

        self._ui.start_chat_menu_button.set_menu_model(get_start_chat_button_menu())

        self._connect(
            self._ui.workspace_settings_button,
            "clicked",
            self._on_edit_workspace_clicked,
        )

        self._ui.paned.set_position(app.settings.get("chat_handle_position"))
        self._ui.paned.connect("notify::position", self._on_handle_position_notify)
        self.toggle_chat_list()

        self._startup_finished: bool = False
        self._closed_chat_memory: list[tuple[str, JID, str]] = []

        self._add_actions()

    def do_unroot(self) -> None:
        self._disconnect_all()
        Gtk.Box.do_unroot(self)

    def _add_actions(self):
        actions = [
            ("remove-chat", "as", self._remove_chat),
            ("search-history", None, self._on_search_history),
        ]

        for action in actions:
            action_name, variant, func = action
            if variant is not None:
                variant = GLib.VariantType.new(variant)
            act = Gio.SimpleAction.new(action_name, variant)
            self._connect(act, "activate", func)
            app.window.add_action(act)

    def set_startup_finished(self) -> None:
        self._startup_finished = True

    def get_chat_list_stack(self) -> ChatListStack:
        return self._chat_list_stack

    def get_chat_stack(self) -> ChatStack:
        return self._chat_stack

    @staticmethod
    def _on_handle_position_notify(paned: Gtk.Paned, *args: Any) -> None:
        position = paned.get_position()
        app.settings.set("chat_handle_position", position)

    def _on_section_label_enter(
        self,
        controller: Gtk.EventControllerMotion,
        _x: int,
        _y: int,
    ) -> None:
        self._ui.workspace_settings_button.set_visible(True)

    def _on_section_label_leave(self, controller: Gtk.EventControllerMotion) -> None:
        self._ui.workspace_settings_button.set_visible(False)

    def _on_edit_workspace_clicked(self, _button: Gtk.Button) -> None:
        app.window.activate_action("win.edit-workspace", GLib.Variant("s", ""))

    def _on_chat_selected(
        self, _chat_list_stack: ChatListStack, workspace_id: str, account: str, jid: JID
    ) -> None:

        self._chat_stack.show_chat(account, jid)

        if (
            not self._search_revealer.get_reveal_child()
            and self._restore_occupants_list
            and self._chat_control.contact.is_groupchat
        ):
            # GroupchatRoster was hidden by Search initially, but Search was
            # afterwards closed in a 1:1 chat. Only restore GroupchatRoster if
            # a group chat was selected.
            app.settings.set("hide_groupchat_occupants_list", False)
            self._restore_occupants_list = False

        if self._search_revealer.get_reveal_child():
            self._search_view.set_context(account, jid)

        self.emit("chat-selected", workspace_id, account, jid)

    def _on_chat_unselected(self, _chat_list_stack: ChatListStack) -> None:
        self._chat_stack.clear()
        self._search_view.set_context(None, None)

    def _on_search_history(
        self, _action: Gio.SimpleAction, _param: Literal[None]
    ) -> None:

        if self._chat_control.has_active_chat():
            self._search_view.set_context(
                self._chat_control.contact.account, self._chat_control.contact.jid
            )

        if not app.settings.get("hide_groupchat_occupants_list"):
            # Hide group chat roster in order to make some space horizontally.
            # Store state to be able to restore it when hiding search.
            self._restore_occupants_list = True
            app.settings.set("hide_groupchat_occupants_list", True)

        self._search_revealer.set_reveal_child(True)
        self._search_view.set_focus()

    def _on_search_hide(self, *args: Any) -> None:
        self.hide_search()

    def _on_chat_list_changed(self, chat_list_stack: ChatListStack, *args: Any) -> None:

        chat_list = chat_list_stack.get_current_chat_list()
        assert chat_list is not None
        name = app.settings.get_workspace_setting(chat_list.workspace_id, "name")
        self._ui.section_label.set_text(name)
        self._ui.search_entry.set_text("")
        self._ui.chat_list_scrolled.get_vadjustment().set_value(0)

    def add_chat_list(self, workspace_id: str) -> None:
        self._chat_list_stack.add_chat_list(workspace_id)

    def show_workspace_chats(self, workspace_id: str) -> None:
        self._chat_list_stack.show_chat_list(workspace_id)
        self._chat_filter.reset()

    def update_workspace(self, workspace_id: str) -> None:
        name = app.settings.get_workspace_setting(workspace_id, "name")
        self._ui.section_label.set_text(name)

    def remove_chat_list(self, workspace_id: str) -> None:
        self._chat_list_stack.remove_chat_list(workspace_id)

    def chat_exists(self, account: str, jid: JID) -> bool:
        return self._chat_list_stack.contains_chat(account, jid)

    def select_chat(self, account: str, jid: JID) -> None:
        self._chat_list_stack.select_chat(account, jid)

    def chat_exists_for_workspace(
        self, workspace_id: str, account: str, jid: JID
    ) -> bool:

        return self._chat_list_stack.contains_chat(
            account, jid, workspace_id=workspace_id
        )

    def add_chat_for_workspace(
        self,
        workspace_id: str,
        account: str,
        jid: JID,
        type_: Literal["chat", "groupchat", "pm"],
        pinned: bool = False,
        position: int = -1,
        select: bool = False,
        message: str | None = None,
    ) -> None:

        client = app.get_client(account)

        if type_ == "chat":
            client.get_module("Contacts").add_chat_contact(jid)

        elif type_ == "groupchat":
            client.get_module("Contacts").add_group_chat_contact(jid)

        elif type_ == "pm":
            client.get_module("Contacts").add_private_contact(jid)

        if self.chat_exists(account, jid):
            if select:
                self._chat_list_stack.select_chat(account, jid)
            return

        self._chat_list_stack.add_chat(
            workspace_id, account, jid, type_, pinned, position
        )

        if self._startup_finished:
            if select:
                self._chat_list_stack.select_chat(account, jid)
            self._chat_list_stack.store_open_chats(workspace_id)
            if message is not None:
                message_input = self._chat_stack.get_message_input()
                message_input.insert_text(message)

    def load_workspace_chats(self, workspace_id: str) -> None:
        open_chats = app.settings.get_workspace_setting(workspace_id, "chats")

        active_accounts = app.settings.get_active_accounts()
        for open_chat in open_chats:
            account = open_chat["account"]
            if account not in active_accounts:
                continue

            self.add_chat_for_workspace(
                workspace_id,
                account,
                open_chat["jid"],
                open_chat["type"],
                pinned=open_chat["pinned"],
                position=open_chat["position"],
            )

    def is_chat_selected(self, account: str, jid: JID) -> bool:
        return self._chat_list_stack.is_chat_selected(account, jid)

    def restore_chat(self) -> None:
        if not self._closed_chat_memory:
            return

        account, jid, workspace_id = self._closed_chat_memory.pop()

        client = app.get_client(account)
        contact = client.get_module("Contacts").get_contact(jid)
        assert isinstance(
            contact, BareContact | GroupchatContact | GroupchatParticipant
        )
        self.add_chat_for_workspace(
            workspace_id, account, jid, contact.type_string, select=True
        )

    def _remove_chat(self, _action: Gio.SimpleAction, param: GLib.Variant) -> None:

        account, jid = param.unpack()
        jid = JID.from_string(jid)

        self.remove_chat(account, jid)

    def remove_chat(self, account: str, jid: JID) -> None:
        for workspace_id in app.settings.get_workspaces():
            if self.chat_exists_for_workspace(workspace_id, account, jid):
                self._closed_chat_memory.append((account, jid, workspace_id))
                self._chat_list_stack.remove_chat(workspace_id, account, jid)
                return

    def _on_chat_removed(
        self,
        _chat_list: ChatList,
        account: str,
        jid: JID,
        type_: str,
        unregister: bool,
    ) -> None:

        if self._chat_control.is_loaded(account, jid):
            self._chat_control.clear()

        if type_ == "groupchat" and app.account_is_connected(account):
            client = app.get_client(account)
            client.get_module("MUC").leave(jid, unregister=unregister)

    def remove_chats_for_account(self, account: str) -> None:
        chat_list = self._chat_list_stack.get_current_chat_list()
        if chat_list is not None:
            chat_list.unselect_all()

        self._chat_list_stack.remove_chats_for_account(account)
        if self._chat_control.has_active_chat():
            if self._chat_control.contact.account == account:
                self._chat_control.clear()

    def get_control(self) -> ChatControl:
        return self._chat_control

    def hide_search(self) -> bool:
        if self._search_revealer.get_reveal_child():
            self._search_revealer.set_reveal_child(False)

            if (
                self._restore_occupants_list
                and self._chat_control.has_active_chat()
                and self._chat_control.contact.is_groupchat
            ):
                # Restore GroupchatRoster only if a group chat is selected
                # currently. If this condition isn' satisfied, it is checked
                # again as soon as a chat is selected in _on_chat_selected.
                app.settings.set("hide_groupchat_occupants_list", False)
                self._restore_occupants_list = False

            return True
        return False

    def toggle_chat_list(self) -> None:
        chat_list = self._ui.paned.get_start_child()
        assert chat_list is not None
        chat_list.set_visible(not chat_list.get_visible())



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

from gi.repository import Gtk

from gajim.common import app
from gajim.common import ged
from gajim.common.client import Client
from gajim.common.const import AvatarSize
from gajim.common.const import ClientState
from gajim.common.events import MucDecline
from gajim.common.events import MucInvitation
from gajim.common.events import SubscribePresenceReceived
from gajim.common.events import UnsubscribedPresenceReceived
from gajim.common.ged import EventHelper
from gajim.common.modules.contacts import BareContact

from gajim.gtk.builder import get_builder
from gajim.gtk.menus import get_account_menu
from gajim.gtk.menus import get_account_notifications_menu
from gajim.gtk.notification_manager import NotificationManager
from gajim.gtk.status_message_selector import StatusMessageSelector
from gajim.gtk.status_selector import StatusSelector
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.window import open_window


class AccountPage(Gtk.Box, EventHelper, SignalManager):
    def __init__(self, account: str) -> None:
        Gtk.Box.__init__(self)
        EventHelper.__init__(self)
        SignalManager.__init__(self)

        self._account = account
        self._client = app.get_client(account)

        jid = self._client.get_own_jid().bare
        self._contact = self._client.get_module("Contacts").get_contact(jid)
        self._contact.connect("avatar-update", self._on_avatar_update)

        self._ui = get_builder("account_page.ui")
        self.append(self._ui.scrolled)

        self._connect(
            self._ui.account_settings_button, "clicked", self._on_account_settings
        )

        self._ui.our_jid_label.set_text(jid)

        self._status_selector = StatusSelector(account=account)
        self._status_selector.set_halign(Gtk.Align.CENTER)
        self._ui.status_box.append(self._status_selector)

        self._status_message_selector = StatusMessageSelector(account=account)
        self._status_message_selector.set_halign(Gtk.Align.CENTER)
        self._ui.status_box.append(self._status_message_selector)

        self._notification_manager = NotificationManager(account)
        self._ui.account_box.append(self._notification_manager)

        self._ui.notifications_menu_button.set_menu_model(
            get_account_notifications_menu(account)
        )

        self._ui.account_page_menu_button.set_menu_model(get_account_menu(account))

        self._client.connect_signal("state-changed", self._on_client_state_changed)

        app.settings.connect_signal(
            "account_label", self._on_account_label_changed, account
        )

        # pylint: disable=line-too-long
        self.register_events(
            [
                ("subscribe-presence-received", ged.GUI1, self._subscribe_received),
                (
                    "unsubscribed-presence-received",
                    ged.GUI1,
                    self._unsubscribed_received,
                ),
                ("muc-invitation", ged.GUI1, self._muc_invitation_received),
                ("muc-decline", ged.GUI1, self._muc_invitation_declined),
            ]
        )
        # pylint: enable=line-too-long

        self.update()

    def do_unroot(self) -> None:
        self._disconnect_all()
        self.unregister_events()
        self._contact.disconnect_all_from_obj(self)
        self._client.disconnect_all_from_obj(self)
        app.settings.disconnect_signals(self)
        Gtk.Box.do_unroot(self)
        app.check_finalize(self)

    def _on_account_label_changed(self, _value: str, *args: Any) -> None:
        self.update()

    def _on_avatar_update(self, *args: Any) -> None:
        self.update()

    def _on_account_settings(self, _button: Gtk.Button) -> None:
        window = open_window("AccountsWindow")
        window.select_account(self._account)

    def _on_client_state_changed(
        self, client: Client, _signal_name: str, state: ClientState
    ) -> None:

        jid = client.get_own_jid().bare
        self._ui.our_jid_label.set_text(jid)

    def update(self) -> None:
        account_label = app.settings.get_account_setting(self._account, "account_label")
        self._ui.account_label.set_text(account_label)

        assert isinstance(self._contact, BareContact)
        texture = self._contact.get_avatar(
            AvatarSize.ACCOUNT_PAGE, self.get_scale_factor(), add_show=False
        )
        self._ui.avatar_image.set_pixel_size(AvatarSize.ACCOUNT_PAGE)
        self._ui.avatar_image.set_from_paintable(texture)

        self._status_selector.update()

    def _subscribe_received(self, event: SubscribePresenceReceived) -> None:
        if event.account != self._account:
            return
        self._notification_manager.add_subscription_request(event)

    def _unsubscribed_received(self, event: UnsubscribedPresenceReceived) -> None:
        if event.account != self._account:
            return
        self._notification_manager.add_unsubscribed(event)

    def _muc_invitation_received(self, event: MucInvitation) -> None:
        if event.account != self._account:
            return
        self._notification_manager.add_invitation_received(event)

    def _muc_invitation_declined(self, event: MucDecline) -> None:
        if event.account != self._account:
            return
        self._notification_manager.add_invitation_declined(event)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast

from gi.repository import Gio
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Pango

from gajim.common import app

GajimDropDownDataT = dict[Any, str] | list[str]


class GajimDropDown(Gtk.DropDown):
    def __init__(
        self,
        data: GajimDropDownDataT | None = None,
        fixed_width: int = -1,
    ) -> None:

        Gtk.DropDown.__init__(self)

        self._kwargs: dict[str, Any] = {
            "width_chars": fixed_width,
            "max_width_chars": fixed_width,
        }

        self._model = Gio.ListStore(item_type=KeyValueItem)
        list_store_expression = Gtk.PropertyExpression.new(
            KeyValueItem,
            None,
            "value",
        )

        self.set_expression(list_store_expression)
        self.set_search_match_mode(Gtk.StringFilterMatchMode.SUBSTRING)
        self.set_model(self._model)

        factory = Gtk.SignalListItemFactory()
        factory.connect("setup", self._on_factory_setup, KeyValueViewItem, self._kwargs)
        factory.connect("bind", self._on_factory_bind)

        self.set_factory(factory)

        factory = Gtk.SignalListItemFactory()
        factory.connect("setup", self._on_factory_setup, KeyValueViewListItem, {})
        factory.connect("bind", self._on_factory_bind)

        self.set_list_factory(factory)

        self.set_data(data)

    @staticmethod
    def _on_factory_setup(
        _factory: Gtk.SignalListItemFactory,
        list_item: Gtk.ListItem,
        view_item: Any,
        kwargs: dict[str, Any],
    ) -> None:
        list_item.set_child(view_item(**kwargs))

    @staticmethod
    def _on_factory_bind(
        _factory: Gtk.SignalListItemFactory, list_item: Gtk.ListItem
    ) -> None:
        view_item = cast(KeyValueViewItem, list_item.get_child())
        obj = cast(KeyValueItem, list_item.get_item())
        view_item.bind(obj)

    def set_data(self, data: GajimDropDownDataT | None) -> None:
        self._model.remove_all()

        if not data:
            return

        items: list[KeyValueItem] = []
        if isinstance(data, dict):
            for key, value in data.items():
                items.append(KeyValueItem(key=key, value=value))

        if isinstance(data, list):
            for entry in data:
                items.append(KeyValueItem(key=entry, value=entry))

        self._model.splice(0, 0, items)

    def get_selected_key(self) -> Any | None:
        selected_item = self.get_selected_item()
        if selected_item is None:
            return None
        return selected_item.get_property("key")

    def select_key(self, key: Any) -> None:
        for pos in range(self._model.get_n_items()):
            item = self._model.get_item(pos)
            assert item is not None
            if item.props.key == key:
                self.set_selected(pos)

    def do_unroot(self) -> None:
        Gtk.DropDown.do_unroot(self)
        self.set_model(None)
        self._model.remove_all()
        app.check_finalize(self._model)
        del self._model
        app.check_finalize(self)


class KeyValueItem(GObject.Object):
    key = GObject.Property(type=object, flags=GObject.ParamFlags.READWRITE)
    value = GObject.Property(type=str, flags=GObject.ParamFlags.READWRITE)


class KeyValueViewItem(Gtk.Label):
    def __init__(self, **kwargs: Any):
        Gtk.Label.__init__(self, ellipsize=Pango.EllipsizeMode.END, xalign=0, **kwargs)

    def bind(self, item: KeyValueItem) -> None:
        self.set_label(item.value)


class KeyValueViewListItem(Gtk.Label):
    def __init__(self, **kwargs: Any):
        Gtk.Label.__init__(self, ellipsize=Pango.EllipsizeMode.END, xalign=0, **kwargs)

    def bind(self, item: KeyValueItem) -> None:
        self.set_label(item.value)



# Copyright (C) 2019 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast

import locale
import logging
import time

from gi.repository import Gtk
from omemo_dr.const import OMEMOTrust
from omemo_dr.structs import IdentityInfo

from gajim.common import app
from gajim.common import ged
from gajim.common import types
from gajim.common.events import AccountConnected
from gajim.common.events import AccountDisconnected
from gajim.common.ged import EventHelper
from gajim.common.helpers import generate_qr_code
from gajim.common.i18n import _
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.modules.contacts import GroupchatParticipant
from gajim.common.modules.omemo import compose_trust_uri

from gajim.gtk.builder import get_builder
from gajim.gtk.dialogs import ConfirmationDialog
from gajim.gtk.dialogs import DialogButton
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import clear_listbox
from gajim.gtk.util.window import open_window

log = logging.getLogger("gajim.gtk.omemo_trust_manager")


TRUST_DATA = {
    OMEMOTrust.UNTRUSTED: ("dialog-error-symbolic", _("Untrusted"), "error-color"),
    OMEMOTrust.UNDECIDED: ("security-low-symbolic", _("Not Decided"), "warning-color"),
    OMEMOTrust.VERIFIED: ("security-high-symbolic", _("Verified"), "encrypted-color"),
    OMEMOTrust.BLIND: ("security-medium-symbolic", _("Blind Trust"), "encrypted-color"),
}


class OMEMOTrustManager(Gtk.Box, EventHelper, SignalManager):
    def __init__(self, account: str, contact: types.ChatContactT | None = None) -> None:
        Gtk.Box.__init__(self)
        EventHelper.__init__(self)
        SignalManager.__init__(self)

        self._account = account
        self._contact = contact

        self._ui = get_builder("omemo_trust_manager.ui")
        self.append(self._ui.stack)

        self._connect(self._ui.search, "search-changed", self._on_search_changed)
        self._connect(
            self._ui.manage_trust_button, "clicked", self._on_manage_trust_clicked
        )
        self._connect(
            self._ui.show_inactive_switch, "notify::active", self._on_show_inactive
        )
        self._connect(
            self._ui.clear_devices_button, "clicked", self._on_clear_devices_clicked
        )

        self._ui.list.set_filter_func(self._filter_func, None)
        self._ui.list.set_sort_func(self._sort_func, None)

        self.register_events(
            [
                ("account-connected", ged.GUI2, self._on_account_state),
                ("account-disconnected", ged.GUI2, self._on_account_state),
            ]
        )

        if not app.account_is_connected(account):
            self._ui.stack.set_visible_child_name("no-connection")
            return

        self._update()

    def do_unroot(self) -> None:
        self.unregister_events()
        self._disconnect_all()
        self._ui.list.set_filter_func(None)
        self._ui.list.set_sort_func(None)

        Gtk.Box.do_unroot(self)
        app.check_finalize(self)

    def _update(self) -> None:
        client = app.get_client(self._account)
        if self._contact is None:
            self._contact = client.get_module("Contacts").get_contact(
                client.get_own_jid().bare
            )

        if isinstance(self._contact, BareContact) and self._contact.is_self:
            header_text = _("Other devices connected with your account")
            popover_qr_text = _(
                "Compare this code with the one shown on your "
                "contact’s screen to ensure the safety of "
                "your end-to-end encrypted chat."
            )
        else:
            assert isinstance(
                self._contact, BareContact | GroupchatContact | GroupchatParticipant
            )
            header_text = _('Devices connected with "%s"') % self._contact.name
            popover_qr_text = (
                _(
                    "Compare this code with the one shown on your "
                    "contact’s screen to ensure the safety of "
                    "your end-to-end encrypted chat "
                    "with %s."
                )
                % self._contact.name
            )

        self._ui.list_heading.set_text(header_text)
        self._ui.comparing_instructions.set_text(popover_qr_text)

        our_fpr_formatted = client.get_module("OMEMO").backend.get_our_fingerprint(
            formatted=True
        )
        self._ui.our_fingerprint_1.set_text(our_fpr_formatted)
        self._ui.our_fingerprint_2.set_text(our_fpr_formatted)

        self.update()
        self._load_qrcode()

    def update(self) -> None:
        assert self._contact is not None
        clear_listbox(self._ui.list)

        if isinstance(self._contact, BareContact) and self._contact.is_self:
            self._ui.clear_devices_button.set_visible(True)
            self._ui.list_heading_box.set_halign(Gtk.Align.START)
        else:
            self._ui.manage_trust_button.set_visible(True)
            if self._contact.is_groupchat:
                self._ui.search_button.set_visible(True)
            else:
                self._ui.list_heading_box.set_halign(Gtk.Align.START)

        assert isinstance(
            self._contact, BareContact | GroupchatContact | GroupchatParticipant
        )
        self._load_fingerprints(self._contact)

    def _on_account_state(self, event: AccountConnected | AccountDisconnected) -> None:

        if event.account != self._account:
            return

        if isinstance(event, AccountConnected):
            self._update()
            self._ui.stack.set_visible_child_name("manage-keys")
        else:
            self._ui.stack.set_visible_child_name("no-connection")

    def _filter_func(self, row: KeyRow, _user_data: Any) -> bool:
        search_text = self._ui.search.get_text()
        if search_text and search_text.lower() not in row.address:
            return False
        if self._ui.show_inactive_switch.get_active():
            return True
        return row.active

    @staticmethod
    def _sort_func(row1: KeyRow, row2: KeyRow, _user_data: Any) -> int:
        result = locale.strcoll(row1.address, row2.address)
        if result != 0:
            return result

        if row1.active != row2.active:
            return -1 if row1.active else 1

        if row1.trust != row2.trust:
            return -1 if row1.trust > row2.trust else 1
        return 0

    def _on_search_changed(self, _entry: Gtk.SearchEntry) -> None:
        self._ui.list.invalidate_filter()

    def _load_fingerprints(self, contact: types.ChatContactT) -> None:
        client = app.get_client(self._account)
        for identity_info in client.get_module("OMEMO").backend.get_identity_infos(
            str(contact.jid)
        ):
            self._ui.list.append(KeyRow(contact, identity_info))

    def _load_qrcode(self) -> None:
        client = app.get_client(self._account)
        uri = compose_trust_uri(
            client.get_own_jid(),
            [client.get_module("OMEMO").backend.get_our_identity()],
        )
        log.debug("Trust URI: %s", uri)
        self._ui.qr_code_image.set_from_paintable(generate_qr_code(uri))

    def _on_show_inactive(self, switch: Gtk.Switch, _param: Any) -> None:
        self._ui.list.invalidate_filter()

    def _on_clear_devices_clicked(self, _button: Gtk.Button) -> None:
        def _clear():
            client = app.get_client(self._account)
            client.get_module("OMEMO").clear_devicelist()

        ConfirmationDialog(
            _("Clear Devices?"),
            _("This will clear the devices store for your account."),
            [
                DialogButton.make("Cancel"),
                DialogButton.make("Accept", text=_("_Clear Devices"), callback=_clear),
            ],
        )

    def _on_manage_trust_clicked(self, _button: Gtk.Button) -> None:
        assert self._contact is not None
        window = open_window("AccountsWindow")
        window.select_account(self._contact.account, "encryption-omemo")


class KeyRow(Gtk.ListBoxRow):
    def __init__(
        self, contact: types.ChatContactT, identity_info: IdentityInfo
    ) -> None:

        Gtk.ListBoxRow.__init__(self)
        self.set_activatable(False)

        self._contact = contact
        self._client = app.get_client(contact.account)

        self._address = identity_info.address
        self._identity_info = identity_info
        self._trust = identity_info.trust

        grid = Gtk.Grid()
        grid.set_column_spacing(12)

        self._trust_button = TrustButton(self)
        grid.attach(self._trust_button, 1, 1, 1, 3)

        if contact.is_groupchat:
            jid_label = Gtk.Label(label=self._address)
            jid_label.set_selectable(False)
            jid_label.set_halign(Gtk.Align.START)
            jid_label.set_valign(Gtk.Align.START)
            jid_label.set_hexpand(True)
            jid_label.add_css_class("bold")
            grid.attach(jid_label, 2, 1, 1, 1)

        self.fingerprint = Gtk.Label(
            label=self._identity_info.public_key.get_fingerprint(formatted=True)
        )
        self.fingerprint.add_css_class("monospace")
        self.fingerprint.add_css_class("small-label")
        self.fingerprint.set_selectable(True)
        self.fingerprint.set_halign(Gtk.Align.START)
        self.fingerprint.set_valign(Gtk.Align.START)
        self.fingerprint.set_hexpand(True)
        grid.attach(self.fingerprint, 2, 2, 1, 1)

        if self._identity_info.last_seen is not None:
            last_seen_str = time.strftime(
                app.settings.get("date_time_format"),
                time.localtime(self._identity_info.last_seen),
            )
        else:
            last_seen_str = _("Never")
        last_seen_label = Gtk.Label(label=_("Last seen: %s") % last_seen_str)
        last_seen_label.set_halign(Gtk.Align.START)
        last_seen_label.set_valign(Gtk.Align.START)
        last_seen_label.set_hexpand(True)
        last_seen_label.add_css_class("small-label")
        last_seen_label.add_css_class("dim-label")
        grid.attach(last_seen_label, 2, 3, 1, 1)

        self.set_child(grid)

    def do_unroot(self) -> None:
        del self._trust_button

        Gtk.ListBoxRow.do_unroot(self)
        app.check_finalize(self)

    def delete_fingerprint(self, *args: Any) -> None:

        def _remove():
            self._client.get_module("OMEMO").backend.delete_session(
                self._address, self._identity_info.device_id, delete_identity=True
            )

            listbox = cast(Gtk.ListBox, self.get_parent())
            listbox.remove(self)

        ConfirmationDialog(
            _("Delete Fingerprint?"),
            _("Doing so will permanently delete this Fingerprint"),
            [
                DialogButton.make("Cancel"),
                DialogButton.make("Remove", text=_("Delete"), callback=_remove),
            ],
        )

    def set_trust(self, trust: OMEMOTrust) -> None:
        self._trust = trust
        icon_name, tooltip, css_class = TRUST_DATA[trust]
        image = cast(Gtk.Image, self._trust_button.get_child())
        image.set_from_icon_name(icon_name)
        image.add_css_class(css_class)
        image.set_tooltip_text(tooltip)

        self._client.get_module("OMEMO").backend.set_trust(
            self._address, self._identity_info.public_key, trust
        )

    @property
    def trust(self) -> OMEMOTrust:
        return self._trust

    @property
    def active(self) -> bool:
        return self._identity_info.active

    @property
    def address(self) -> str:
        return self._address


class TrustButton(Gtk.MenuButton):
    def __init__(self, row: KeyRow) -> None:
        Gtk.MenuButton.__init__(self)
        self._row = row
        self._css_class = ""
        self._trust_popover = TrustPopver(row, self)
        self.set_popover(self._trust_popover)
        self.set_valign(Gtk.Align.CENTER)
        self.update()

    def do_unroot(self) -> None:
        del self._trust_popover
        del self._row

        Gtk.MenuButton.do_unroot(self)
        app.check_finalize(self)

    def update(self) -> None:
        icon_name, tooltip, css_class = TRUST_DATA[self._row.trust]
        image = Gtk.Image.new_from_icon_name(icon_name)
        self.set_child(image)

        if not self._row.active:
            css_class = "omemo-inactive-color"
            tooltip = f'{_("Inactive")} - {tooltip}'

        image.add_css_class(css_class)
        self._css_class = css_class
        self.set_tooltip_text(tooltip)


class TrustPopver(Gtk.Popover, SignalManager):
    def __init__(self, row: KeyRow, trust_button: TrustButton) -> None:
        Gtk.Popover.__init__(self)
        SignalManager.__init__(self)
        self._row = row
        self._trust_button = trust_button

        self._listbox = Gtk.ListBox()
        self._listbox.set_selection_mode(Gtk.SelectionMode.NONE)
        self.update()
        self.set_child(self._listbox)
        self._connect(self._listbox, "row-activated", self._activated)
        self.add_css_class("omemo-trust-popover")

    def do_unroot(self) -> None:
        Gtk.Popover.do_unroot(self)
        self._disconnect_all()
        del self._listbox
        del self._row
        del self._trust_button
        app.check_finalize(self)

    def _activated(self, _listbox: Gtk.ListBox, row: MenuOption) -> None:
        self.popdown()
        if row.type_ is None:
            self._row.delete_fingerprint()
        else:
            self._row.set_trust(row.type_)
            self._trust_button.update()
            self.update()

    def update(self) -> None:
        clear_listbox(self._listbox)

        if self._row.trust != OMEMOTrust.VERIFIED:
            self._listbox.append(VerifiedOption())
        if self._row.trust != OMEMOTrust.BLIND:
            self._listbox.append(BlindOption())
        if self._row.trust != OMEMOTrust.UNTRUSTED:
            self._listbox.append(NotTrustedOption())
        self._listbox.append(DeleteOption())


class MenuOption(Gtk.ListBoxRow):
    def __init__(
        self, icon: str, label_text: str, color: str, type_: OMEMOTrust | None = None
    ) -> None:

        Gtk.ListBoxRow.__init__(self)

        self.type_ = type_
        self.icon = icon
        self.label = label_text
        self.color = color

        box = Gtk.Box()
        box.set_spacing(6)

        image = Gtk.Image.new_from_icon_name(icon)
        if color:
            image.add_css_class(color)

        label = Gtk.Label(label=label_text)

        box.append(image)
        box.append(label)
        self.set_child(box)


class BlindOption(MenuOption):
    def __init__(self) -> None:
        MenuOption.__init__(
            self,
            "security-medium-symbolic",
            _("Blind Trust"),
            "encrypted-color",
            OMEMOTrust.BLIND,
        )


class VerifiedOption(MenuOption):
    def __init__(self) -> None:
        MenuOption.__init__(
            self,
            "security-high-symbolic",
            _("Verified"),
            "encrypted-color",
            OMEMOTrust.VERIFIED,
        )


class NotTrustedOption(MenuOption):
    def __init__(self) -> None:
        MenuOption.__init__(
            self,
            "dialog-error-symbolic",
            _("Untrusted"),
            "error-color",
            OMEMOTrust.UNTRUSTED,
        )


class DeleteOption(MenuOption):
    def __init__(self) -> None:
        MenuOption.__init__(self, "user-trash-symbolic", _("Delete"), "")



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from typing import Any
from typing import cast
from typing import NamedTuple

import logging
from datetime import timedelta

from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import Gtk
from gi.repository import Pango
from nbxmpp.errors import MalformedStanzaError
from nbxmpp.errors import StanzaError
from nbxmpp.modules.dataforms import DataForm
from nbxmpp.namespaces import Namespace
from nbxmpp.structs import LastActivityData
from nbxmpp.structs import SoftwareVersionResult
from nbxmpp.task import Task

from gajim.common import app
from gajim.common import ged
from gajim.common.const import TLS_VERSION_STRINGS
from gajim.common.events import ServerDiscoReceived
from gajim.common.ged import EventHelper
from gajim.common.i18n import _
from gajim.common.util.uri import open_uri

from gajim.gtk.builder import get_builder
from gajim.gtk.certificate_dialog import CertificateBox
from gajim.gtk.widgets import GajimAppWindow

log = logging.getLogger("gajim.gtk.server_info")


class Feature(NamedTuple):
    name: str
    available: bool
    additional: str | None = None


class ServerInfo(GajimAppWindow, EventHelper):
    def __init__(self, account: str) -> None:
        GajimAppWindow.__init__(
            self,
            name="ServerInfo",
            title=_("Server Info"),
            default_width=600,
            default_height=700,
        )
        EventHelper.__init__(self)

        self.account = account
        self._client = app.get_client(account)

        if app.settings.get("use_kib_mib"):
            self._units = GLib.FormatSizeFlags.IEC_UNITS
        else:
            self._units = GLib.FormatSizeFlags.DEFAULT

        self._ui = get_builder("server_info.ui")
        self.set_child(self._ui.server_info_notebook)

        self._connect(
            self._ui.clipboard_button, "clicked", self._on_clipboard_button_clicked
        )

        self.register_events(
            [
                ("server-disco-received", ged.GUI1, self._server_disco_received),
            ]
        )

        self._version = ""
        self._hostname = app.get_hostname_from_account(account, prefer_custom=True)
        self._ui.server_hostname.set_text(self._hostname)

        domain = self._client.get_own_jid().domain
        self._client.get_module("SoftwareVersion").request_software_version(
            domain, callback=self._software_version_received
        )

        self._client.get_module("LastActivity").request_last_activity(
            domain, callback=self._on_last_activity
        )

        server_info = self._client.get_module("Discovery").server_info
        self._add_contact_addresses(server_info.dataforms)

        self._add_connection_info()

        if self._client.certificate is None:
            self._ui.no_certificate_label.set_visible(True)
        else:
            cert_box = CertificateBox(account, self._client.certificate)
            self._ui.cert_scrolled.set_child(cert_box)
            self._ui.cert_scrolled.set_visible(True)

        for feature in self._get_features():
            self._add_feature(feature)

    def _cleanup(self, *args: Any) -> None:
        self.unregister_events()

    def _add_connection_info(self) -> None:
        # Connection type
        nbxmpp_client = self._client.connection
        address = nbxmpp_client.current_address

        assert address is not None
        self._ui.connection_type.set_text(address.type.value)
        if address.type.is_plain:
            self._ui.connection_type.add_css_class("error-color")

        # Connection proxy
        proxy = address.proxy
        if proxy is not None:
            self._ui.proxy_type.set_text(proxy.type)
            self._ui.proxy_host.set_text(proxy.host)

        self._ui.domain.set_text(address.domain)

        visible = address.service is not None
        self._ui.dns_label.set_visible(visible)
        self._ui.dns.set_visible(visible)
        self._ui.dns.set_text(address.service or "")

        visible = nbxmpp_client.remote_address is not None
        self._ui.ip_port_label.set_visible(visible)
        self._ui.ip_port.set_visible(visible)
        self._ui.ip_port.set_text(nbxmpp_client.remote_address or "")

        visible = address.uri is not None
        self._ui.websocket_label.set_visible(visible)
        self._ui.websocket.set_visible(visible)
        self._ui.websocket.set_text(address.uri or "")

        assert nbxmpp_client is not None
        tls_version = TLS_VERSION_STRINGS.get(nbxmpp_client.tls_version)
        self._ui.tls_version.set_text(tls_version or _("Not available"))

        self._ui.cipher_suite.set_text(nbxmpp_client.ciphersuite or _("Not available"))

    def _add_contact_addresses(self, dataforms: list[DataForm]) -> None:
        fields = {
            "admin-addresses": _("Admin"),
            "status-addresses": _("Status"),
            "support-addresses": _("Support"),
            "security-addresses": _("Security"),
            "feedback-addresses": _("Feedback"),
            "abuse-addresses": _("Abuse"),
            "sales-addresses": _("Sales"),
        }

        addresses = self._get_addresses(fields, dataforms)
        if addresses is None:
            self._ui.no_addresses_label.set_visible(True)
            return

        row_count = 4
        for address_type, values in addresses.items():
            label = self._get_address_type_label(fields[address_type])
            self._ui.server.attach(label, 0, row_count, 1, 1)
            for index, value in enumerate(values):
                last = index == len(values) - 1
                label = self._get_address_label(value, last=last)
                self._ui.server.attach(label, 1, row_count, 1, 1)
                row_count += 1

    @staticmethod
    def _get_addresses(
        fields: dict[str, str], dataforms: list[DataForm]
    ) -> dict[str, list[str]] | None:
        addresses: dict[str, list[str]] = {}
        for form in dataforms:
            field = form.vars.get("FORM_TYPE")
            if field.value != "http://jabber.org/network/serverinfo":
                continue

            for address_type in fields:
                field = form.vars.get(address_type)
                if field is None:
                    continue

                if field.type_ != "list-multi":
                    continue

                if not field.values:
                    continue
                addresses[address_type] = field.values

            return addresses or None
        return None

    @staticmethod
    def _get_address_type_label(text: str) -> Gtk.Label:
        label = Gtk.Label(label=text)
        label.set_halign(Gtk.Align.END)
        label.set_valign(Gtk.Align.START)
        label.add_css_class("dim-label")
        return label

    def _get_address_label(self, address: str, last: bool = False) -> Gtk.Label:
        label = Gtk.Label()
        label.set_markup(f'<a href="{address}">{address}</a>')
        label.set_ellipsize(Pango.EllipsizeMode.END)
        label.set_xalign(0)
        label.set_halign(Gtk.Align.START)
        label.add_css_class("link-button")
        self._connect(label, "activate-link", self._on_activate_link)
        if last:
            label.set_margin_bottom(6)
        return label

    def _on_activate_link(self, label: Gtk.Label, *args: Any) -> int:
        open_uri(label.get_text(), account=self.account)
        return Gdk.EVENT_STOP

    def _on_last_activity(self, task: Task) -> None:
        try:
            result = cast(LastActivityData, task.finish())
        except (StanzaError, MalformedStanzaError) as error:
            log.warning(error)
            return

        delta = timedelta(seconds=result.seconds)
        hours = 0
        if result.seconds >= 3600:
            hours = delta.seconds // 3600
        uptime = _("%(days)s days, %(hours)s hours") % {
            "days": delta.days,
            "hours": hours,
        }
        self._ui.server_uptime.set_text(uptime)

    def _software_version_received(self, task: Task) -> None:
        try:
            result = cast(SoftwareVersionResult, task.finish())
        except StanzaError:
            self._version = _("Unknown")
        else:
            self._version = f"{result.name} {result.version}"

        self._ui.server_software.set_text(self._version)

    def _server_disco_received(self, _event: ServerDiscoReceived) -> None:
        features = self._get_features()
        for index, item in enumerate(features):
            row = cast(FeatureItem, self._ui.features_listbox.get_row_at_index(index))
            row.update(item)

    def _add_feature(self, feature: Feature) -> None:
        item = FeatureItem(feature)
        self._ui.features_listbox.append(item)

    def _get_features(self) -> list[Feature]:
        http_upload_module = self._client.get_module("HTTPUpload")
        http_upload_info = http_upload_module.httpupload_namespace
        if http_upload_module.available:
            max_size = http_upload_module.max_file_size
            if max_size is not None:
                max_size = GLib.format_size_full(int(max_size), self._units)
                http_upload_info = f"{http_upload_info} (max. {max_size})"

        return [
            Feature(
                "XEP-0045: Multi-User Chat", self._client.get_module("MUC").supported
            ),
            Feature(
                "XEP-0054: vcard-temp", self._client.get_module("VCardTemp").supported
            ),
            Feature(
                "XEP-0077: In-Band Registration",
                self._client.get_module("Register").supported,
            ),
            Feature(
                "XEP-0163: Personal Eventing Protocol",
                self._client.get_module("PEP").supported,
            ),
            Feature(
                "XEP-0163: #publish-options",
                self._client.get_module("PubSub").publish_options,
            ),
            Feature(
                "XEP-0191: Blocking Command",
                self._client.get_module("Blocking").supported,
                Namespace.BLOCKING,
            ),
            Feature(
                "XEP-0198: Stream Management",
                self._client.features.has_sm(),
                Namespace.STREAM_MGMT,
            ),
            Feature(
                "XEP-0258: Security Labels in XMPP",
                self._client.get_module("SecLabels").supported,
                Namespace.SECLABEL,
            ),
            Feature(
                "XEP-0280: Message Carbons",
                self._client.get_module("Carbons").supported,
                Namespace.CARBONS,
            ),
            Feature(
                "XEP-0313: Message Archive Management",
                self._client.get_module("MAM").available,
            ),
            Feature(
                "XEP-0363: HTTP File Upload",
                self._client.get_module("HTTPUpload").available,
                http_upload_info,
            ),
            Feature(
                "XEP-0398: Avatar Conversion",
                self._client.get_module("VCardAvatars").avatar_conversion_available,
            ),
            Feature(
                "XEP-0411: Bookmarks Conversion",
                self._client.get_module("Bookmarks").conversion,
            ),
            Feature(
                "XEP-0402: Bookmarks Compat",
                self._client.get_module("Bookmarks").compat,
            ),
            Feature(
                "XEP-0402: Bookmarks Compat PEP",
                self._client.get_module("Bookmarks").compat_pep,
            ),
        ]

    def _on_clipboard_button_clicked(self, _widget: Gtk.Button) -> None:
        server_software = _("Server Software: %s\n") % self._version
        server_features = ""

        for feature in self._get_features():
            if feature.available:
                available = _("Available")
            else:
                available = _("Not available")
            additional = ""
            if feature.additional is not None:
                additional = f"({feature.additional})"
            server_features += f"{feature.name}: {available} {additional}\n"

        self.window.get_clipboard().set(server_software + server_features)


class FeatureItem(Gtk.ListBoxRow):
    def __init__(self, feature: Feature) -> None:
        Gtk.ListBoxRow.__init__(self)
        self._feature = feature

        grid = Gtk.Grid(row_spacing=3, column_spacing=12)

        self._icon = Gtk.Image()
        self._feature_label = Gtk.Label()
        self._feature_label.set_halign(Gtk.Align.START)
        self._additional_label = Gtk.Label()
        self._additional_label.set_halign(Gtk.Align.START)
        self._additional_label.set_visible(False)
        self._additional_label.add_css_class("dim-label")

        grid.attach(self._icon, 0, 0, 1, 1)
        grid.attach(self._feature_label, 1, 0, 1, 1)
        grid.attach(self._additional_label, 1, 1, 1, 1)

        self._set_feature()
        self.set_child(grid)

    def _set_feature(self) -> None:
        self._feature_label.set_text(self._feature.name)
        if self._feature.additional is not None:
            self._additional_label.set_text(self._feature.additional)
            self._additional_label.set_visible(True)

        self._icon.remove_css_class("error-color")
        self._icon.remove_css_class("success-color")

        if self._feature.available:
            self._icon.set_from_icon_name("emblem-ok-symbolic")
            self._icon.add_css_class("success-color")
        else:
            self._icon.set_from_icon_name("window-close-symbolic")
            self._icon.add_css_class("error-color")

    def update(self, feature: Feature) -> None:
        self._feature = feature
        self._set_feature()



# Copyright (C) 2014 Kamil Paral <kamil.paral AT gmail.com>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

'''
Watch for system shutdown using systemd-logind.
Documentation: https://www.freedesktop.org/wiki/Software/systemd/inhibit
'''
from __future__ import annotations

from typing import Any

import logging
import os

from gi.repository import Gio
from gi.repository import GLib

from gajim.common import app
from gajim.common.i18n import _

log = logging.getLogger('gajim.c.dbus.logind')


class LogindListener:
    _instance: LogindListener | None = None

    @classmethod
    def get(cls) -> LogindListener:
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def __init__(self) -> None:
        # file descriptor object of the inhibitor
        self._inhibit_fd: int | None = None

        Gio.bus_watch_name(
            Gio.BusType.SYSTEM,
            'org.freedesktop.login1',
            Gio.BusNameWatcherFlags.NONE,
            self._on_logind_appeared,
            self._on_logind_vanished)

    def _on_prepare_for_shutdown(self,
                                 _connection: Gio.DBusConnection,
                                 _sender_name: str,
                                 _object_path: str,
                                 interface_name: str,
                                 signal_name: str,
                                 parameters: tuple[str, str, str],
                                 *_user_data: Any
                                 ) -> None:
        '''Signal handler for PrepareForShutdown event'''
        log.debug('Received signal %s.%s%s',
                  interface_name, signal_name, parameters)

        if self._inhibit_fd is None:
            log.warning('Preparing for shutdown by quitting Gajim, '
                        'without holding a shutdown inhibitor')
        else:
            log.info('Preparing for shutdown by quitting Gajim')

        app.window.quit()

    def _obtain_delay_inhibitor(self, connection: Gio.DBusConnection) -> None:
        '''Obtain a shutdown delay inhibitor from logind'''
        if self._inhibit_fd is not None:
            # Something is wrong, we have an inhibitor fd, and we are asking for
            # yet another one.
            log.warning('Trying to obtain a shutdown inhibitor '
                        'while already holding one.')
            return

        try:
            result = connection.call_with_unix_fd_list_sync(
                'org.freedesktop.login1',
                '/org/freedesktop/login1',
                'org.freedesktop.login1.Manager',
                'Inhibit',
                GLib.Variant('(ssss)', (
                    'shutdown',
                    app.get_default_app_id(),
                    _('Shutting down Gajim'),
                    'delay'  # Inhibitor will delay but not block shutdown
                )),
                GLib.VariantType.new('(h)'),
                Gio.DBusCallFlags.NONE,
                -1,
                None,
                None)
        except GLib.Error as error:
            log.warning(
                'Could not obtain a shutdown delay inhibitor from '
                'logind: %s', error)
            return

        ret, ret_fdlist = result
        if ret_fdlist is None:  # pyright: ignore
            # This can happen as reported by users
            log.warning('Unable to obtain inhibitor, fdlist is None')
            return

        self._inhibit_fd = ret_fdlist.get(ret.unpack()[0])
        log.info('Obtained shutdown delay inhibitor')

    def release_delay_inhibitor(self) -> None:
        '''Release our shutdown delay inhibitor'''
        if self._inhibit_fd is not None:
            os.close(self._inhibit_fd)
            self._inhibit_fd = None
        log.info('Released shutdown delay inhibitor')

    def _on_logind_appeared(self,
                            connection: Gio.DBusConnection,
                            name: str,
                            name_owner: str,
                            *_user_data: Any
                            ) -> None:
        '''Use signal and locks provided by org.freedesktop.login1'''
        log.info('Name %s appeared, owned by %s', name, name_owner)

        connection.signal_subscribe(
            'org.freedesktop.login1',
            'org.freedesktop.login1.Manager',
            'PrepareForShutdown',
            '/org/freedesktop/login1',
            None,
            Gio.DBusSignalFlags.NONE,
            self._on_prepare_for_shutdown,
            None)
        self._obtain_delay_inhibitor(connection)

    def _on_logind_vanished(self,
                            _connection: Gio.DBusConnection,
                            name: str,
                            *_user_data: Any
                            ) -> None:
        '''Release remaining resources related to org.freedesktop.login1'''
        log.info('Name %s vanished', name)
        self.release_delay_inhibitor()


def enable() -> None:
    LogindListener.get()


def shutdown() -> None:
    LogindListener.get().release_delay_inhibitor()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from typing import Any
from typing import cast

import logging

from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import Gtk
from gi.repository import Pango

from gajim.common import app
from gajim.common.i18n import _
from gajim.common.preview import Preview
from gajim.common.util.preview import contains_audio_streams
from gajim.common.util.preview import format_geo_coords
from gajim.common.util.preview import get_icon_for_mime_type
from gajim.common.util.preview import split_geo_uri

from gajim.gtk.builder import get_builder
from gajim.gtk.menus import get_preview_menu
from gajim.gtk.preview_audio import AudioWidget
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import ensure_not_destroyed
from gajim.gtk.widgets import GajimPopover

log = logging.getLogger("gajim.gtk.preview")

PREVIEW_ACTIONS: dict[str, tuple[str, str]] = {
    "open": (_("Open"), "win.preview-open"),
    "save_as": (_("Save as…"), "win.preview-save-as"),
    "open_folder": (_("Open Folder"), "win.preview-open-folder"),
    "copy_link_location": (_("Copy Link"), "win.preview-copy-link"),
    "open_link_in_browser": (_("Open Link in Browser"), "win.preview-open-link"),
    "download": (_("Download File"), "win.preview-download"),
}


class PreviewWidget(Gtk.Box, SignalManager):
    def __init__(self, account: str) -> None:
        Gtk.Box.__init__(self)
        SignalManager.__init__(self)

        self.account = account
        self._preview: Preview | None = None

        self._destroyed = False

        if app.settings.get("use_kib_mib"):
            self._units = GLib.FormatSizeFlags.IEC_UNITS
        else:
            self._units = GLib.FormatSizeFlags.DEFAULT

        self._ui = get_builder("preview.ui")
        self.append(self._ui.preview_stack)

        self._connect(self._ui.icon_button, "clicked", self._on_content_button_clicked)
        self._connect(
            self._ui.cancel_download_button, "clicked", self._on_cancel_download_clicked
        )
        self._connect(self._ui.image_button, "clicked", self._on_content_button_clicked)

        content_hover_controller = Gtk.EventControllerMotion()
        self._connect(content_hover_controller, "enter", self._on_content_cursor_enter)
        self._connect(content_hover_controller, "leave", self._on_content_cursor_leave)
        self._ui.preview_box.add_controller(content_hover_controller)

        self._connect(self._ui.open_folder_button, "clicked", self._on_open_folder)
        self._connect(self._ui.save_as_button, "clicked", self._on_save_as)
        self._connect(self._ui.download_button, "clicked", self._on_download)

        pointer_cursor = Gdk.Cursor.new_from_name("pointer")
        self._ui.icon_button.set_cursor(pointer_cursor)
        self._ui.cancel_download_button.set_cursor(pointer_cursor)
        self._ui.image_button.set_cursor(pointer_cursor)
        self._ui.download_button.set_cursor(pointer_cursor)
        self._ui.save_as_button.set_cursor(pointer_cursor)
        self._ui.open_folder_button.set_cursor(pointer_cursor)

        self._menu_popover = GajimPopover(None)
        self.append(self._menu_popover)

        leftclick_action = app.settings.get("preview_leftclick_action")
        self._ui.icon_button.set_tooltip_text(PREVIEW_ACTIONS[leftclick_action][0])

        gesture_secondary_click = Gtk.GestureClick(button=Gdk.BUTTON_SECONDARY)
        self._connect(gesture_secondary_click, "pressed", self._on_preview_clicked)
        self.add_controller(gesture_secondary_click)

        app.settings.connect_signal(
            "preview_leftclick_action", self._update_icon_button_tooltip
        )

    def do_unroot(self) -> None:
        self._destroyed = True
        self._disconnect_all()
        app.settings.disconnect_signals(self)
        self._preview = None
        del self._preview
        Gtk.Box.do_unroot(self)
        app.check_finalize(self)

    def _update_icon_button_tooltip(self, setting: str, *args: Any) -> None:
        self._ui.icon_button.set_tooltip_text(PREVIEW_ACTIONS[setting][0])

    def get_text(self) -> str:
        if self._preview is None:
            return ""
        return self._preview.uri

    @ensure_not_destroyed
    def update_progress(self, preview: Preview, progress: float) -> None:
        self._preview = preview

        self._ui.preview_stack.set_visible_child_name("preview")
        self._ui.download_button.set_visible(False)

        self._ui.progress_box.set_visible(True)
        self._ui.progress_text.set_label(f"{int(progress * 100)} %")
        self._ui.progressbar.set_fraction(progress)
        self._ui.info_message.set_text(_("Downloading…"))
        self._ui.info_message.set_tooltip_text("")

    @ensure_not_destroyed
    def update(self, preview: Preview, data: bytes | None) -> None:
        self._preview = preview

        self._ui.preview_stack.set_visible_child_name("preview")
        self._ui.progress_box.set_visible(False)
        self._ui.info_message.set_visible(False)

        if preview.is_geo_uri:
            image = Gtk.Image.new_from_gicon(Gio.ThemedIcon(name="map"))
            image.set_pixel_size(preview.size)
            self._ui.image_button.set_child(image)

            self._ui.icon_event_box.set_visible(False)
            self._ui.file_name.set_text(_("Click to view location"))
            self._ui.file_name.set_selectable(False)
            self._ui.save_as_button.set_visible(False)
            self._ui.open_folder_button.set_visible(False)
            self._ui.download_button.set_visible(False)

            location = split_geo_uri(preview.uri)
            text = format_geo_coords(float(location.lat), float(location.lon))
            self._ui.file_size.set_text(text)
            self._ui.image_button.set_tooltip_text(_("Location at %s") % text)
            self._ui.preview_box.set_size_request(160, -1)
            return

        self._ui.image_button.set_tooltip_text(preview.filename)

        if data is not None:
            self._ui.image_button.set_tooltip_text(None)

            texture = Gdk.Texture.new_from_bytes(GLib.Bytes.new(data))

            max_preview_size = app.settings.get("preview_size")

            texture_width = texture.get_width()
            texture_height = texture.get_height()

            if texture_width > max_preview_size or texture_height > max_preview_size:
                # Scale down with or height to max_preview_size
                if texture_width > texture_height:
                    width = max_preview_size
                    height = int(max_preview_size / texture_width * texture_height)
                else:
                    width = int(max_preview_size / texture_height * texture_width)
                    height = max_preview_size
            else:
                width = texture_width
                height = texture_height

            # Set minimum height of 100 to avoid button overlay covering entire image
            self._ui.content_overlay.set_size_request(width, max(height, 100))

            image = Gtk.Picture.new_for_paintable(texture)
            image.add_css_class("preview-image")
            image.set_can_target(False)
            self._ui.content_overlay.add_overlay(image)

            self._ui.button_box.unparent()
            self._ui.content_overlay.add_overlay(self._ui.button_box)
            self._ui.button_box.set_visible(False)

            self._ui.button_box.set_valign(Gtk.Align.END)
            self._ui.button_box.set_halign(Gtk.Align.FILL)
            self._ui.button_box.set_can_target(True)

            self._ui.preview_stack.remove_css_class("preview-stack")
            self._ui.preview_stack.add_css_class("preview-stack-image")
            self._ui.button_box.add_css_class("preview-image-overlay")
            self._ui.open_folder_button.add_css_class("preview-image-overlay-button")
            self._ui.save_as_button.add_css_class("preview-image-overlay-button")

        else:
            icon = get_icon_for_mime_type(preview.mime_type)
            image = Gtk.Image.new_from_gicon(icon)
            image.set_pixel_size(64)
            self._ui.icon_button.set_child(image)

        preview_enabled = app.settings.get("enable_file_preview")

        if preview_enabled and preview.is_previewable and preview.orig_exists:
            self._ui.icon_event_box.set_visible(False)
            self._ui.image_button.set_visible(True)
            self._ui.save_as_button.set_visible(True)
            self._ui.open_folder_button.set_visible(True)
        else:
            self._ui.image_button.set_visible(False)
            self._ui.icon_event_box.set_visible(True)

        file_size_string = _("File size unknown")
        if preview.file_size != 0:
            file_size_string = GLib.format_size_full(preview.file_size, self._units)

        self._ui.link_button.set_uri(preview.uri)
        self._ui.link_button.set_tooltip_text(preview.uri)
        self._ui.link_button.set_label(preview.uri)
        label = cast(Gtk.Label, self._ui.link_button.get_child())
        label.set_ellipsize(Pango.EllipsizeMode.END)
        label.set_max_width_chars(32)

        if preview.info_message is not None:
            self._ui.info_message.set_text(preview.info_message)
            self._ui.info_message.set_tooltip_text(preview.info_message)
            self._ui.info_message.set_visible(True)

        if preview.orig_exists:
            if preview_enabled:
                self._ui.link_button.set_visible(False)

            self._ui.download_button.set_visible(False)
            self._ui.open_folder_button.set_visible(True)
            self._ui.save_as_button.set_visible(True)

            if (
                preview_enabled
                and preview.orig_path is not None
                and preview.is_audio
                and app.is_installed("GST")
                and contains_audio_streams(preview.orig_path)
            ):
                self._ui.image_button.set_visible(False)
                audio_widget = AudioWidget(preview.orig_path)
                self._ui.right_box.append(audio_widget)
                self._ui.right_box.reorder_child_after(
                    audio_widget, self._ui.content_box
                )
        else:
            if preview.file_size == 0:
                if preview_enabled:
                    self._ui.download_button.set_visible(False)
                else:
                    self._ui.download_button.set_visible(True)
                self._ui.link_button.set_visible(True)
            else:
                self._ui.download_button.set_visible(True)
            self._ui.save_as_button.set_visible(False)
            self._ui.open_folder_button.set_visible(False)
            allow_in_public = app.settings.get("preview_anonymous_muc")
            if (
                preview.context == "public"
                and not allow_in_public
                and not preview.from_us
            ):
                image = Gtk.Image.new_from_icon_name("dialog-question")
                image.set_pixel_size(64)
                self._ui.icon_button.set_child(image)
                self._ui.download_button.set_visible(True)
                file_size_string = _("Automatic preview disabled")

        self._ui.file_size.set_text(file_size_string)
        self._ui.file_name.set_text(preview.filename)
        self._ui.file_name.set_tooltip_text(preview.filename)

    def _on_download(self, _button: Gtk.Button) -> None:
        if self._preview is None:
            return

        variant = GLib.Variant("s", self._preview.id)
        app.window.activate_action("win.preview-download", variant)

    def _on_save_as(self, _button: Gtk.Button) -> None:
        assert self._preview is not None
        variant = GLib.Variant("s", self._preview.id)
        app.window.activate_action("win.preview-save-as", variant)

    def _on_open_folder(self, _button: Gtk.Button) -> None:
        assert self._preview is not None
        variant = GLib.Variant("s", self._preview.id)
        app.window.activate_action("win.preview-open-folder", variant)

    def _on_content_button_clicked(self, _button: Gtk.Button) -> None:
        if self._preview is None:
            return

        leftclick_action = app.settings.get("preview_leftclick_action")
        variant = GLib.Variant("s", self._preview.id)
        action = PREVIEW_ACTIONS[leftclick_action][1]
        app.window.activate_action(action, variant)

    def _on_preview_clicked(
        self,
        _gesture_click: Gtk.GestureClick,
        _n_press: int,
        x: float,
        y: float,
    ) -> None:
        if self._preview is None:
            return

        menu = get_preview_menu(self._preview)
        self._menu_popover.set_menu_model(menu)
        self._menu_popover.set_pointing_to_coord(x, y)
        self._menu_popover.popup()

    def _on_content_cursor_enter(
        self,
        _controller: Gtk.EventControllerMotion,
        _x: int,
        _y: int,
    ) -> None:
        assert self._preview is not None
        if self._preview.mime_type.startswith("image/"):
            self._ui.button_box.set_visible(True)

    def _on_content_cursor_leave(
        self,
        _controller: Gtk.EventControllerMotion,
    ) -> None:
        assert self._preview is not None
        if self._preview.mime_type.startswith("image/") and self._preview.orig_exists:
            self._ui.button_box.set_visible(False)

    def _on_cancel_download_clicked(self, _button: Gtk.Button) -> None:
        assert self._preview is not None
        app.preview_manager.cancel_download(self._preview)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# Message Util module

from __future__ import annotations

from datetime import datetime
from datetime import UTC

from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.structs import EMEData
from nbxmpp.structs import MessageProperties

from gajim.common.const import EME_MESSAGES
from gajim.common.modules.contacts import GroupchatParticipant
from gajim.common.storage.archive import models as mod
from gajim.common.storage.archive.const import ChatDirection
from gajim.common.storage.archive.const import MessageType
from gajim.common.storage.base import VALUE_MISSING
from gajim.common.structs import MUCData


def get_eme_message(eme_data: EMEData) -> str:
    try:
        return EME_MESSAGES[eme_data.namespace]
    except KeyError:
        return EME_MESSAGES['fallback'] % eme_data.name


def get_chat_type_and_direction(
    muc_data: MUCData | None, own_jid: JID, properties: MessageProperties
) -> tuple[MessageType, ChatDirection]:

    assert properties.jid is not None
    if properties.type.is_groupchat:
        assert muc_data is not None
        direction = ChatDirection.INCOMING
        if muc_data.occupant_id is not None:
            if muc_data.occupant_id == properties.occupant_id:
                direction = ChatDirection.OUTGOING

        elif muc_data.nick == properties.jid.resource:
            direction = ChatDirection.OUTGOING

        return MessageType.GROUPCHAT, direction

    assert properties.from_ is not None
    if properties.from_.bare_match(own_jid):
        direction = ChatDirection.OUTGOING
    else:
        direction = ChatDirection.INCOMING

    if properties.is_muc_pm:
        return MessageType.PM, direction

    return MessageType.CHAT, direction


def get_real_jid(
    properties: MessageProperties,
    contact: GroupchatParticipant,
) -> JID | None:

    if properties.is_mam_message:
        if properties.muc_user is None:
            return None
        return properties.muc_user.jid

    real_jid = contact.real_jid
    if real_jid is None:
        return None
    return real_jid.new_as_bare()


def get_occupant_info(
    account: str,
    remote_jid: JID,
    own_bare_jid: JID,
    direction: ChatDirection,
    timestamp: datetime,
    contact: GroupchatParticipant,
    properties: MessageProperties
) -> mod.Occupant | None:

    assert properties.jid is not None
    if properties.jid.is_bare:
        return None

    if direction == ChatDirection.OUTGOING:
        real_jid = own_bare_jid
    else:
        real_jid = get_real_jid(properties, contact)

    occupant_id = get_occupant_id(contact, properties) or real_jid
    if occupant_id is None:
        return None

    resource = properties.jid.resource
    assert resource is not None

    return mod.Occupant(
        account_=account,
        remote_jid_=remote_jid,
        id=str(occupant_id),
        real_remote_jid_=real_jid or VALUE_MISSING,
        nickname=resource,
        updated_at=timestamp,
    )


def get_occupant_id(
    contact: GroupchatParticipant,
    properties: MessageProperties
) -> str | None:

    if not properties.occupant_id:
        return None

    if contact.room.supports(Namespace.OCCUPANT_ID):
        return properties.occupant_id
    return None


def get_message_timestamp(
    properties: MessageProperties
) -> datetime:
    timestamp = properties.timestamp
    if properties.mam is not None:
        timestamp = properties.mam.timestamp
    return datetime.fromtimestamp(timestamp, tz=UTC)


def convert_message_type(type_: MessageType) -> str:
    if type_ in (MessageType.CHAT, MessageType.PM):
        return 'chat'
    return 'groupchat'


def get_nickname_from_message(message: mod.Message) -> str:
    if message.resource is None:
        nickname = message.remote.jid.localpart
        assert nickname is not None
        return nickname

    if message.occupant is not None:
        nickname = message.occupant.nickname
        assert nickname is not None
        return nickname

    nickname = message.resource
    assert nickname is not None
    return nickname
